import os
import tempfile
import shutil
import subprocess
import multiprocessing

import pyranges as pr
import pandas as pd

import handygenome
import handygenome.deco as deco
import handygenome.refgenome.refgenome as refgenome
import handygenome.logutils as logutils
from handygenome.genomedf.genomedf import GenomeDataFrame as GDF
from handygenome.cnv.depth import DepthRawDataFrame as DepthRawDF
import handygenome.tools as tools
import handygenome.bameditor as bameditor


def get_mosdepth_args(
    prefix, bam_path, t, use_median, no_perbase=True, bed_path=None, window_size=None,
):
    if (
        (bed_path is not None)
        and (window_size is not None)
    ):
        raise Exception(f'"window_size" and "bed_path" arguments must not be used at the same time.')

    args = [handygenome.PARAMS['mosdepth'], '-x', '-t', str(t)]

    if use_median:
        args.append('-m')

    if no_perbase:
        args.append('-n')

    if bed_path is not None:
        args.extend(('-b', bed_path))
    elif window_size is not None:
        args.extend(('-b', str(window_size)))

    args.extend([prefix, bam_path])

    return args


@deco.get_deco_num_set_differently(('region_bed_path', 'region_gdf'), 2, how='lt')
def run_mosdepth(
    bam_path, 
    #refver=None,
    t=1, 
    nproc=1,
    use_median=False, 

    region_bed_path=None, 
    region_gdf=None, 

    window=None,

    verbose=True,
    prefix='PREFIX'
):
    # set refver
    #if refver is None:
    refver = refgenome.infer_refver_bampath(bam_path)

    # make bam index
    bam_idx_path = bameditor.get_index_path(bam_path)
    if bam_idx_path is None:
        _ = pysam.index(bam_path)

    region_gdf_list = run_mosdepth_prepare_region_gdfs(
        region_bed_path=region_bed_path,
        region_gdf=region_gdf,
        window=window,
        refver=refver,
        n_split=nproc,
    )

    # run parallel jobs
    tmpdir = tempfile.mkdtemp(dir=os.getcwd(), prefix='mosdepth_tmpdir_')
    bam_basename = os.path.basename(bam_path)
    if verbose:
        logutils.log(f'Running mosdepth parallel jobs (bam: {bam_basename})', add_locstring=False)

    args = (
        (
            bam_path, 
            region_gdf,
            refver,
            t, 
            use_median, 
            prefix,
            False,  # "verbose" argument
            tmpdir,
        )
        for region_gdf in region_gdf_list
    )
    with multiprocessing.Pool(nproc) as pool:
        pool_result = pool.starmap(run_mosdepth_base, args)

    shutil.rmtree(tmpdir)

    if verbose:
        logutils.log(f'Finished mosdepth parallel jobs (bam: {bam_basename})', add_locstring=False)

    result = DepthRawDF.concat(pool_result)

    return result


def run_mosdepth_prepare_region_gdfs(
    region_bed_path,
    region_gdf,
    window,
    refver,
    n_split,
):
    region_is_given = (region_bed_path is not None) or (region_gdf is not None)
    if region_is_given:
        if region_bed_path is not None:
            region_gdf = GDF.read_tsv(region_bed_path, refver)
    else:
        region_gdf = GDF.all_regions(refver, assembled_only=False)

    if window is not None:
        region_gdf = region_gdf.window(window)

    region_gdf.sort()
    region_gdf_list = region_gdf.equal_nrow_split(n=n_split)
    return region_gdf_list


#@deco.get_deco_num_set_differently(('region_bed_path', 'region_gdf'), 1)
def run_mosdepth_base(
    bam_path, 
    region_gdf,
    refver,
    t=8, 
    use_median=False, 
    prefix='PREFIX',
    verbose=False,
    tmpdir_location=None,
):
    """- Input bam must be indexed.
    - region bed file: May be gzipped. May not have header row.
    """
    # make tmp directory
    if tmpdir_location is None:
        tmpdir_location = os.getcwd()
    tmpdir = tempfile.mkdtemp(dir=tmpdir_location, prefix='mosdepth_tmpdir_')

    # write region bed file
    if region_gdf is None:
        mosdepth_input_bed_path = None
    else:
        mosdepth_input_bed_path = os.path.join(tmpdir, 'region.bed')
        region_gdf.drop_annots().df.to_csv(
            mosdepth_input_bed_path, 
            sep='\t', 
            header=False, 
            index=False,
        )

    # prepare mosdepth input bam
    mosdepth_input_bam_path = os.path.join(tmpdir, 'mosdepth_input.bam')
    os.symlink(bam_path, mosdepth_input_bam_path)

    # make index of the symlink
    bam_idx_path = bameditor.get_index_path(bam_path)
    if bam_idx_path is None:
        _ = pysam.index(mosdepth_input_bam_path)
    else:
        os.symlink(bam_idx_path, mosdepth_input_bam_path + '.bai')

    # make mosdepth outdir
    outdir = os.path.join(tmpdir, 'outdir')
    os.mkdir(outdir)

    # run mosdepth
    bam_basename = os.path.basename(bam_path)
    if verbose:
        logutils.log(f'Running mosdepth (bam: {bam_basename})', add_locstring=False)
    mosdepth_args = get_mosdepth_args(
        prefix=os.path.join(outdir, prefix), 
        bam_path=mosdepth_input_bam_path, 
        t=t, 
        use_median=use_median, 
        no_perbase=True,
        bed_path=mosdepth_input_bed_path, 
    )
    p = subprocess.check_call(mosdepth_args)
    if verbose:
        logutils.log(f'Finished running mosdepth (bam: {bam_basename})', add_locstring=False)

    # load mosdepth outputs
    outfile_path = os.path.join(outdir, f'{prefix}.regions.bed.gz')
    result_gdf = DepthRawDF.load_mosdepth(outfile_path, refver, use_median=use_median)

    # remove tmpdir
    shutil.rmtree(tmpdir)

    # return
    return result_gdf


def run_mosdepth_old(
    bam_path, 
    refver=None,
    t=8, 
    use_median=False, 
    region_bed_path=None, 
    region_gdf=None, 
    window_size=None, 
    #donot_subset_bam=False, 
    subset_bam=None, 
    load_perbase=False,
    prefix='PREFIX'
):
    """- Input bam must be indexed.
    - region bed file: May be gzipped. May not have header row.
    """
    # sanity check
    if (
        (window_size is not None)
        and (
            (region_bed_path is not None)
            or (region_gdf is not None)
        )
    ):
        raise Exception(f'"window_size" and "region*" arguments must not be used at the same time.')

    if (
        (region_bed_path is not None)
        and (region_gdf is not None)
    ):  
        raise Exception(f'Only one of "region_bed_path" and "region_gdf" argument must be set.')

    # make tmp directory
    tmpdir = tempfile.mkdtemp(dir=os.getcwd(), prefix='mosdepth_tmpdir_')

    # set refver
    if refver is None:
        refver = refgenome.infer_refver_bampath(bam_path)

    # write region bed file
    region_is_given = (region_bed_path is not None) or (region_gdf is not None)
    if region_is_given:
        mosdepth_input_bed_path = os.path.join(tmpdir, 'region.bed')
        if region_bed_path is not None:
            region_gdf = GDF.read_tsv(region_bed_path, refver)
        region_gdf.df.iloc[:, :3].to_csv(
            mosdepth_input_bed_path, 
            sep='\t', 
            header=False, 
            index=False,
        )
    else:
        mosdepth_input_bed_path = None

    # prepare mosdepth input bam
    mosdepth_input_bam_path = os.path.join(tmpdir, 'mosdepth_input.bam')
    if subset_bam is None:
        subset_bam = load_perbase
    if (region_is_given and subset_bam): 
        # subset input bam file
        logutils.log(f'Subsetting input bam file', add_locstring=False)
        bameditor.samtools_view(
            in_bam_path=bam_path, 
            out_bam_path=mosdepth_input_bam_path, 
            region_bed_path=mosdepth_input_bed_path, 
            index=True,
        )
    else:  
        # do not subset bam
        os.symlink(bam_path, mosdepth_input_bam_path)

        # make index of the symlink
        bam_idx_path = bameditor.get_index_path(bam_path)
        if bam_idx_path is None:
            _ = pysam.index(mosdepth_input_bam_path)
        else:
            os.symlink(bam_idx_path, mosdepth_input_bam_path + '.bai')

    # make mosdepth outdir
    outdir = os.path.join(tmpdir, 'outdir')
    os.mkdir(outdir)

    # run mosdepth
    logutils.log(f'Running mosdepth', add_locstring=False)
    mosdepth_args = get_mosdepth_args(
        prefix=os.path.join(outdir, prefix), 
        bam_path=mosdepth_input_bam_path, 
        t=t, 
        use_median=use_median, 
        no_perbase=(not load_perbase),
        bed_path=mosdepth_input_bed_path, 
        window_size=window_size,
    )
    p = subprocess.check_call(mosdepth_args)

    # load mosdepth outputs
    region_outfile_path = os.path.join(outdir, f'{prefix}.regions.bed.gz')
    region_result_gdf = DepthRawDF.load_mosdepth(region_outfile_path, refver, use_median=use_median)

    # load mosdepth outputs - perbase
    perbase_outfile_path = os.path.join(outdir, f'{prefix}.per-base.bed.gz')
    if load_perbase:
        perbase_result_gdf = DepthRawDF.load_mosdepth_perbase(perbase_outfile_path, refver)
        if region_is_given:
            perbase_result_gdf = perbase_result_gdf.intersect(region_gdf)
    else:
        perbase_result_gdf = None

    # remove tmpdir
    shutil.rmtree(tmpdir)

    # return
    if load_perbase:
        return region_result_gdf, perbase_result_gdf
    else:
        return region_result_gdf





import os
import functools
import re
import itertools

import pandas as pd
import pyranges as pr

import handygenome
import handygenome.network as network
import handygenome.cnv.misc as cnvmisc
import handygenome.refgenome.refgenome as refgenome
from handygenome.genomedf.genomedf import GenomeDataFrame as GDF


URL_BASE = 'https://api.genome.ucsc.edu'

CYTOBAND_URLS = refgenome.RefverDict({
    'GRCh37': 'http://hgdownload.cse.ucsc.edu/goldenpath/hg19/database/cytoBand.txt.gz', 
    'GRCh38': 'http://hgdownload.cse.ucsc.edu/goldenpath/hg38/database/cytoBand.txt.gz', 
})

CYTOBAND_DIR = os.path.join(handygenome.USERDATA_DIR, 'cytoband')
os.makedirs(CYTOBAND_DIR, exist_ok=True)


def tohexcode(tup):
    return '#' + ''.join(hex(x)[2:] for x in tup)

# https://www.wikiwand.com/en/Chromosome_1
CYTOBAND_COLORMAP = {
    #'gvar': tohexcode((224, 224, 224)),
    'gvar': 'tab:pink',
    'stalk': tohexcode((112, 128, 144)),
    #'acen': tohexcode((110, 127, 143)),
    'acen': 'tab:red',
    'gneg': tohexcode((255, 255, 255)),
    'gpos25': tohexcode((217, 217, 217)),
    'gpos50': tohexcode((151, 151, 151)),
    'gpos75': tohexcode((99, 99, 99)),
    'gpos100': tohexcode((0, 0, 0)),
}


def get_avaliable_genomes():
    url = f'{URL_BASE}/list/ucscGenomes'
    raw = network.http_get(url)
    return raw['ucscGenomes']


@functools.cache
def get_refver_aliases():
    result = dict()
    genome_data = get_avaliable_genomes()
    for key, val in genome_data.items():
        if val['organism'] in ('Mouse', 'Human'):
            aliases = re.match('.*\((\S+ )?(\S+)\)', val['description']).group(2).split('/')
            result[key] = aliases

    return result


def standardize_refver_ucscapi(refver):
    for standard, aliases in get_refver_aliases().items():
        if refver in aliases:
            return standard
    raise Exception(f'Unknown reference version.')


def list_tracks(refver):
    refver = standardize_refver_ucscapi(refver)
    raw = network.http_get(f'{URL_BASE}/list/tracks?genome={refver}')
    return sorted(raw[refver].keys())


def rename_hg19_cytoband(cytoband_df):
    result = cytoband_df.copy()
    assemblyspec = refgenome.SPECS['hg19']
    new_chroms = result['Chromosome'].apply(
        lambda x: assemblyspec.convert(x, 'nochr_plus_genbank')
    )
    result['Chromosome'] = new_chroms
    result = result.loc[result['Chromosome'].notna(), :]

    return result


#def get_cytoband_from_ucsc_api(refver, rename_hg19=True, as_gr=True):
def get_cytoband_from_ucsc_api(refver):
    ucsc_refver = standardize_refver_ucscapi(refver)
    if 'cytoBand' not in list_tracks(ucsc_refver):
        raise Exception(f'"cytoBand" is not available for this reference genome.')

    raw = network.http_get(
        f'{URL_BASE}/getData/track?genome={ucsc_refver};track=cytoBand'
    )
    data = list()
    for dic in itertools.chain.from_iterable(raw['cytoBand'].values()):
        data.append(
            (dic['chrom'], dic['chromStart'], dic['chromEnd'], dic['name'], dic['gieStain'])
        )
    df = pd.DataFrame.from_records(
        data, columns=['Chromosome', 'Start', 'End', 'Name', 'Stain'],
    )

    # remove chromosome names absent from chrom converter
    chromconv = refgenome.get_chrom_converter(refver)
    valid_indexes = df['Chromosome'].isin(tuple(chromconv.keys()))
    df = df.loc[valid_indexes, :]
    df['Chromosome'] = df['Chromosome'].apply(lambda x: chromconv[x])

    # remove chromosome names absent from chromdict
    chromdict = refgenome.get_chromdict(refver)
    valid_indexes = df['Chromosome'].isin(tuple(chromdict.keys()))
    df = df.loc[valid_indexes, :]

    result = GDF.from_frame(df, refver=refver)
    result.sort()
    #if (refver == 'hg19') and rename_hg19:
    #    result = rename_hg19_cytoband(result)

    #result = cnvmisc.sort_genome_df(result, refver=refver)
    #if as_gr:
    #    result = pr.PyRanges(result)

    return result


def get_cytoband_path(refver):
    refver = refgenome.standardize(refver)
    return os.path.join(CYTOBAND_DIR, f'{refver}.tsv.gz')


def write_cytoband(cytoband_df, refver):
    cytoband_df.to_csv(
        get_cytoband_path(refver),
        sep='\t',
        header=True,
        index=False,
    )


def get_cytoband(refver):
    cytoband_path = get_cytoband_path(refver)
    if not os.path.exists(cytoband_path):
        cytoband_gdf = get_cytoband_from_ucsc_api(refver)
        cytoband_gdf.write_tsv(cytoband_path)
        #write_cytoband(cytoband_df, refver)
    else:
        cytoband_gdf = GDF.read_tsv(
            cytoband_path, refver, dtype={'Name': 'string', 'Stain': 'string'},
        )

    return cytoband_gdf
        

get_cytoband_gr = get_cytoband



import base64

import handygenome.network as network
import handygenome.deco as deco


def get_cosmic_auth_string(email, password):
    auth_string = base64.b64encode(f'{email}:{password}'.encode('utf-8')).decode()
    return auth_string


@deco.get_deco_arg_choices({'refver': ('GRCh37', 'GRCh38')})
def get_cosmic_file_url(auth_string, refver, cosmicver, cosmicfile):
    open_url = f'https://cancer.sanger.ac.uk/cosmic/file_download/{refver}/cosmic/{cosmicver}/{cosmicfile}'
    url = network.http_get(open_url, headers={'Authorization' : f'Basic {auth_string}'})['url']
    return url


def download_cosmic_file(outfile_path, email, password, cosmicfile, cosmicver, refver):
    """Args:
        refver: "GRCh37" or "GRCh38"
        cosmicver: e.g. "v95"
        cosmicfile: e.g. "CosmicNCV.tsv.gz"
    """
    auth_string = get_cosmic_auth_string(email, password)
    cosmicfile_url = get_cosmic_file_url(auth_string, refver, cosmicver, cosmicfile)
    network.download(cosmicfile_url, outfile_path)



import re
import os
import argparse

import pysam

import importlib
top_package_name = __name__.split('.')[0]
common = importlib.import_module('.'.join([top_package_name, 'common']))
workflow = importlib.import_module('.'.join([top_package_name, 'workflow']))
toolsetup = importlib.import_module('.'.join([top_package_name, 'workflow', 'toolsetup']))
concat_module = importlib.import_module('.'.join([top_package_name, 'vcfeditor', 'concat']))
indexing = importlib.import_module('.'.join([top_package_name, 'vcfeditor', 'indexing']))


def argument_parser(cmdargs):
    def sanity_check(args):
        pass

    parser_dict = workflow.init_parser()
    workflow.add_infilelist_arg(parser_dict['required'], required=True)
    workflow.add_outfile_arg(parser_dict['required'], required=True, 
                             must_not_exist='ask')
    workflow.add_outfmt_arg(parser_dict['optional'], required=False)

    workflow.add_logging_args(parser_dict)
    workflow.add_index_arg(parser_dict)

    args = parser_dict['main'].parse_args(cmdargs)
    sanity_check(args)

    return args


def main(cmdargs):
    args = argument_parser(cmdargs)
    logger = toolsetup.setup_logger(args)

    logger.info('BEGINNING')

    concat_module.main(infile_path_list=args.infile_path_list,
                       outfile_path=args.outfile_path,
                       mode_pysam=args.mode_pysam, 
                       outfile_must_not_exist='no')
    if not args.donot_index:
        indexing.index_vcf(args.outfile_path)

    logger.info('ALL SUCCESSFULLY FINISHED')



import re

import handygenome.tools as tools
import handygenome.network as network
import handygenome.deco as deco
import handygenome.refgenome.refgenome as refgenome


PREFIX_LOOKUP_ID = refgenome.RefverDict({
    'GRCh37': 'http://grch37.rest.ensembl.org/lookup/id',
    'GRCh38': 'http://rest.ensembl.org/lookup/id',
})

PREFIX_LOOKUP_SYMBOL = refgenome.RefverDict({
    'GRCh37': 'http://grch37.rest.ensembl.org/lookup/symbol',
    'GRCh38': 'http://rest.ensembl.org/lookup/symbol',
})

PREFIX_REGULATORY = refgenome.RefverDict({
    'GRCh37': 'http://grch37.rest.ensembl.org/regulatory/species/homo_sapiens/id',
    'GRCh38': 'http://rest.ensembl.org/regulatory/species/homo_sapiens/id',
})

PREFIX_VEP = refgenome.RefverDict({
    'GRCh37': 'http://grch37.rest.ensembl.org/vep/human/hgvs',
    'GRCh38': 'http://rest.ensembl.org/vep/human/hgvs',
    'GRCm39': 'http://rest.ensembl.org/vep/mouse/hgvs',
})

PREFIX_OVERLAP = refgenome.RefverDict({
    'GRCh37': 'http://grch37.rest.ensembl.org/overlap/region',
    'GRCh38': 'http://rest.ensembl.org/overlap/region',
})

PREFIX_MAP = refgenome.RefverDict({
    'GRCh37': 'http://grch37.rest.ensembl.org/map',
    'GRCh38': 'http://rest.ensembl.org/map',
})

PREFIX_SEQUENCE = refgenome.RefverDict({
    'GRCh37': 'http://grch37.rest.ensembl.org/sequence/id',
    'GRCh38': 'http://rest.ensembl.org/sequence/id',
    'GRCm39': 'http://rest.ensembl.org/sequence/id',
})

HTTP_HEADERS_POST = {
    'Content-Type': 'application/json',
    'Accept': 'application/json',
}

HTTP_HEADERS_GET = {
    'Content-Type': 'application/json',
    'Accept': 'application/json',
}

MAX_POST_SIZE = {
    'lookup_id': 1000,
    'lookup_symbol': 1000,
    'vep': 200,
    'sequence': 50,
}


def lookup_id(ID, refver, expand=False):
    """start, end are 1-based inclusive coordinates"""
    prefix = PREFIX_LOOKUP_ID[refver]
    url = '/'.join([prefix, ID])
    params = {'expand': int(expand)}

    return network.http_get(url, params=params, headers=HTTP_HEADERS_GET)


def lookup_id_post(ID_list, refver, expand=False):
    """start, end are 1-based inclusive coordinates"""
    result = list()

    url = PREFIX_LOOKUP_ID[refver]
    params = {'expand' : int(expand)}
    for sublist in tools.chunk_iter(ID_list, MAX_POST_SIZE['lookup_id']):
        data = {'ids' : sublist}
        subresult = network.http_post(url, data, params=params, headers=HTTP_HEADERS_POST)
        result.extend(subresult.values())

    return result


def lookup_symbol(symbol, refver, expand=False):
    prefix = PREFIX_LOOKUP_SYMBOL[refver]
    species = refgenome.RefverDict.find_species(refver)
    url = '/'.join([prefix, species, symbol])
    params = {'expand' : int(expand)}

    return network.http_get(url, params=params, headers=HTTP_HEADERS_GET)


def lookup_symbol_post(symbols, refver, expand=False):
    prefix = PREFIX_LOOKUP_SYMBOL[refver]
    species = refgenome.RefverDict.find_species(refver)
    url = '/'.join([prefix, species])
    params = {'expand' : int(expand)}

    result = list()
    for sublist in tools.chunk_iter(symbols, MAX_POST_SIZE['lookup_symbol']):
        data = {'symbols' : sublist}
        subresult = network.http_post(url, data, params=params, headers=HTTP_HEADERS_POST)
        result.extend(subresult.values())

    return result


def regulatory(ID, refver, activity=True):
    prefix = PREFIX_REGULATORY[refver]
    url = '/'.join([prefix, ID])
    params = {'activity' : int(activity)}

    result = network.http_get(url, params=params, headers=HTTP_HEADERS_GET)
    if isinstance(result, list) and len(result) == 1:
        result = result[0]
    else:
        raise Exception(f'REST regulatory result for ID {ID} is not a list with length of 1.')

    return result


def vep(
    refver, hgvsg=None, vcfspec=None, distance=5000, 
    with_CADD=True, with_Phenotypes=False, with_canonical=True, 
    with_mane=True, with_miRNA=False, with_numbers=True, 
    with_protein=True, with_ccds=True, with_hgvs=True,
):
    if hgvsg is None:
        if vcfspec is None: 
            raise Exception(f'When "hgvsg" is not set, "vcfspec" must be set.')
        vcfspec.check_monoalt(raise_with_false=True)
        hgvsg = vcfspec.to_hgvsg(alt_index=0)

    prefix = PREFIX_VEP[refver]
    url = '/'.join([prefix, hgvsg])

    params = {
        'distance' : distance,
        'CADD' : int(with_CADD),
        'canonical' : int(with_canonical),
        'mane' : int(with_mane),
        'miRNA' : int(with_miRNA),
        'numbers' : int(with_numbers),
        'protein' : int(with_protein),
        'ccds' : int(with_ccds),
        'hgvs' : int(with_hgvs),
    }
    if with_Phenotypes:
        # Adding 'Phenotypes' to params results in phenotype annotation
            # even when it is set to False
        params.update({'Phenotypes' : int(with_Phenotypes)})

    return network.http_get(url, params=params, headers=HTTP_HEADERS_GET)


def vep_post(
    refver, vcfspec_list=None, hgvsg_list=None, distance=5000,
    with_CADD=True, with_Phenotypes=False, with_canonical=True,
    with_mane=True, with_miRNA=False, with_numbers=True, 
    with_protein=True, with_ccds=True, with_hgvs=True,
):
    url = PREFIX_VEP[refver]
    if hgvsg_list is None:
        if vcfspec_list is None: 
            raise Exception(
                f'When "hgvsg_list" is not set, "vcfspec_list" must be set.'
            )
        if not all(
            vcfspec.check_monoalt(raise_with_false=False)
            for vcfspec in vcfspec_list
        ):
            raise Exception(f'Components of "vcfspec_list" must be all monoalt.')
        hgvsg_list = [vcfspec.to_hgvsg(alt_index=0) for vcfspec in vcfspec_list]

    params = {
        'distance': distance,
        'CADD': int(with_CADD),
        'canonical': int(with_canonical),
        'mane': int(with_mane),
        'miRNA': int(with_miRNA),
        'numbers': int(with_numbers),
        'protein': int(with_protein),
        'ccds': int(with_ccds),
        'hgvs': int(with_hgvs),
    }
    if with_Phenotypes:
        params.update({'Phenotypes': int(with_Phenotypes)})

    result = list()
    for sublist in tools.chunk_iter(hgvsg_list, MAX_POST_SIZE['vep']):
        data = {'hgvs_notations': sublist}
        result.extend(network.http_post(url, data, params=params, headers=HTTP_HEADERS_POST))

    return result


def overlap(
    chrom, start1, end1, refver,
    transcript=True, regulatory=True, motif=False, repeat=False,
):
    """start1, end1: 1-based closed system"""

    prefix = PREFIX_OVERLAP[refver]
    species = refgenome.RefverDict.find_species(refver)

    suffix_list = list()
    if transcript: 
        suffix_list.append('feature=transcript')
    if regulatory: 
        suffix_list.append('feature=regulatory')
    if motif: 
        suffix_list.append('feature=motif')
    if repeat: 
        suffix_list.append('feature=repeat')

    url = '/'.join([prefix, species, f'{chrom}:{start1}-{end1}'])
    if len(suffix_list) > 0:
        url += '?' + ';'.join(suffix_list)

    return network.http_get(url, headers=HTTP_HEADERS_GET)


@deco.get_deco_arg_choices({'mode': ('cdna', 'cds', 'translation')})
def map(ID, start1, end1, mode, refver):
    prefix = PREFIX_MAP[refver]
    url = '/'.join([prefix, mode, ID, f'{start1}..{end1}'])
    return network.http_get(url, headers=HTTP_HEADERS_GET)
    

def sequence_get(ID, refver, type='genomic'):
    prefix = PREFIX_SEQUENCE[refver]
    url = '/'.join([prefix, ID])
    params = {
        'type': type,
    }

    return network.http_get(
        url, 
        params=params, 
        headers={'Content-Type': 'text/plain'},
        text=True,
    )


def sequence_post(ID_list, refver, type='genomic'):
    url = PREFIX_SEQUENCE[refver]
    params = {
        'type': type,
    }

    result = list()
    for sublist in tools.chunk_iter(ID_list, MAX_POST_SIZE['sequence']):
        data = {'ids': sublist}
        result.extend(
            network.http_post(
                url, 
                data, 
                params=params, 
                headers=HTTP_HEADERS_POST,
            )
        )

    return result






#!/home/users/pjh/tools/miniconda/221104/miniconda3/envs/genome_v7/bin/python
import sys
import os
import stat
import re
import textwrap
import shutil

PYTHON = sys.executable
PROJECT_PATH = os.path.dirname(os.path.dirname(os.path.abspath(sys.argv[0])))
PACKAGE_PATH = os.path.join(PROJECT_PATH, 'handygenome')
TARGETDIR = os.path.join(PROJECT_PATH, 'bin')


def main():
    # delete existing ones
    if os.path.exists(TARGETDIR):
        shutil.rmtree(TARGETDIR)
    os.mkdir(TARGETDIR)

    # create new ones
    for top, dirs, files in os.walk(os.path.join(PACKAGE_PATH, 'bin')):
        for fname in files:
            if fname.endswith('.py') and not fname.startswith('_'):
                target_name = os.path.join(TARGETDIR, re.sub('.py$', '', fname))

                abspath = os.path.join(top, fname)
                nopy = re.sub('.py$', '', abspath)
                strip_leading = re.sub(PROJECT_PATH + '/', '', nopy)
                modulepath = '.'.join(strip_leading.split('/'))

                text = textwrap.dedent(
                    f"""\
                    #!{PYTHON}
                    import sys
                    sys.path.append('{os.path.dirname(PACKAGE_PATH)}')
                    from {modulepath} import main
                    main(sys.argv[1:])"""
                )
                with open(target_name, 'wt') as outfile:
                    outfile.write(text)

                os.chmod(
                    target_name, 
                    (
                        stat.S_IRUSR 
                        | stat.S_IWUSR
                        | stat.S_IXUSR
                        | stat.S_IRGRP
                        | stat.S_IXGRP
                    )
                )


if __name__ == '__main__':
    main()



import re
import pprint
import sys
import collections
import copy
import json
import functools
import textwrap

import pysam
import numpy as np

import handygenome.tools as tools
import handygenome.variant.infoformat as infoformat


# annotitem codec functions #

ENCODING_SEPS = ("\x1c", "\x1d")
DEFAULT_SEPS = (",", ":")
PERCENT_TABLE = {
    "%": "%25",
    ":": "%3A",
    ";": "%3B",
    "=": "%3D",
    ",": "%2C",
}
TR_TABLE_DECODER = str.maketrans(dict(zip(ENCODING_SEPS, DEFAULT_SEPS)))


def _replace(string, tup):
    return string.replace(*tup)


def _encode_replacer_percent(string):
    return functools.reduce(_replace, PERCENT_TABLE.items(), string)


def _decode_replacer_percent(string):
    return functools.reduce(
        _replace, (reversed(x) for x in PERCENT_TABLE.items()), string
    )


def _decode_replacer_seps(string):
    return string.translate(TR_TABLE_DECODER)


def _convert_key_to_nonstr(dic):
    result = dict()
    for k, v in dic.items():
        new_k = tools.str_to_nonstr(k)
        if isinstance(v, dict):
            result[new_k] = _convert_key_to_nonstr(v)
        else:
            result[new_k] = v
    return result


def _encode(data, handle_percent=True):
    if handle_percent:
        return _encode_replacer_percent(json.dumps(data, separators=ENCODING_SEPS))
    else:
        return json.dumps(data, separators=ENCODING_SEPS)


def _decode(string, handle_percent=True, convert_key=True):
    if handle_percent:
        json_loaded = json.loads(_decode_replacer_seps(_decode_replacer_percent(string)))
    else:
        json_loaded = json.loads(_decode_replacer_seps(string))

    if convert_key:
        return _convert_key_to_nonstr(json_loaded)
    else:
        return json_loaded

#############################


# CONSTANTS
# SEPS = {
#    "nonprint": {"top": "\x1c", "keyval": "\x1d", "seq": "\x1e", "subkeyval": "\x1f"},
#    "initial": {"top": "$$$", "keyval": ":::", "seq": "&&&", "subkeyval": "@@@"},
#    "percent": {"top": "%3A", "keyval": "%3B", "seq": "%3D", "subkeyval": "%25"},
# }
#DEFAULT_SEPTYPE = "nonprint"

# SEP_TOP = '$$$'
# SEP_KEYVAL = ':::'
# SEP_SEQ = '&&&'
# SEP_SUBKEYVAL = '@@@'

# SEP_TOP = '%3A'
# SEP_KEYVAL = '%3B'
# SEP_SEQ = '%3D'
# SEP_SUBKEYVAL = '%25'

# SEP_TOP = '\x1c'
# SEP_KEYVAL = '\x1d'
# SEP_SEQ = '\x1e'
# SEP_SUBKEYVAL = '\x1f'

#ATOMIC_TYPES = (type(None), str, int, float, bool)


#class AnnotDB:
#    """
#    Attributes:
#        featureDB
#        transcript
#        transcript_canonical
#        regulatory
#        motif
#        repeat
#        popfreq
#        cosmic
#    """
#
#    def __init__(
#        self, annotdb_type, refver, fasta, chromdict, vr=None, septype=DEFAULT_SEPTYPE
#    ):
#        assert (
#            annotdb_type in ANNOTDB_TYPES
#        ), f"annotdb_type must be one of: {ANNOTDB_TYPES}"
#
#        self.annotdb_type = annotdb_type
#        self.refver = refver
#        # assert self.refver in ensembl_rest.REFVERS_ALLOWED
#        self.fasta = fasta
#        self.chromdict = chromdict
#        self.septype = septype
#
#        if vr is None:
#            self.init_empty()
#        else:
#            self.load_vr(vr)
#
#    def __repr__(self):
#        return tools.cpformat(self.get_showdict())
#
#    def get_showdict(self):
#        showdict = dict()
#        showdict["popfreq"] = dict(self.popfreq)
#        showdict["cosmic"] = dict(self.cosmic)
#
#        showdict["transcript"] = self.transcript.get_showdict()
#        showdict["regulatory"] = self.regulatory.get_showdict()
#        showdict["motif"] = self.motif.get_showdict()
#        showdict["repeat"] = self.repeat.get_showdict()
#
#        return showdict
#
#    def init_empty(self):
#        self.transcript = AnnotItemDict(septype=self.septype)
#        self.regulatory = AnnotItemDict(septype=self.septype)
#        self.motif = AnnotItemDict(septype=self.septype)
#        self.repeat = AnnotItemDict(septype=self.septype)
#
#        self.set_transcript_subsets()
#
#        self.popfreq = Popfreq(septype=self.septype)
#        self.cosmic = Cosmic(septype=self.septype)
#
#    def load_vr(self, vr):
#        self.transcript = AnnotItemDict(
#            vr,
#            infokey=INFOMETAS[self.annotdb_type]["transcript"]["ID"],
#            septype=self.septype,
#        )
#        self.regulatory = AnnotItemDict(
#            vr,
#            infokey=INFOMETAS[self.annotdb_type]["regulatory"]["ID"],
#            septype=self.septype,
#        )
#        self.motif = AnnotItemDict(
#            vr,
#            infokey=INFOMETAS[self.annotdb_type]["motif"]["ID"],
#            septype=self.septype,
#        )
#        self.repeat = AnnotItemDict(
#            vr,
#            infokey=INFOMETAS[self.annotdb_type]["repeat"]["ID"],
#            septype=self.septype,
#        )
#
#        self.set_transcript_subsets()
#
#        self.popfreq = Popfreq(
#            vr,
#            infokey=INFOMETAS[self.annotdb_type]["popfreq"]["ID"],
#            septype=self.septype,
#        )
#        self.cosmic = Cosmic(
#            vr,
#            infokey=INFOMETAS[self.annotdb_type]["cosmic"]["ID"],
#            septype=self.septype,
#        )
#
#    def write(self, vr, add_meta=False):
#        if add_meta:
#            add_infometas(vr.header)
#
#        self.transcript.write_info(
#            vr, key=INFOMETAS[self.annotdb_type]["transcript"]["ID"], add_meta=False
#        )
#        self.regulatory.write_info(
#            vr, key=INFOMETAS[self.annotdb_type]["regulatory"]["ID"], add_meta=False
#        )
#        self.motif.write_info(
#            vr, key=INFOMETAS[self.annotdb_type]["motif"]["ID"], add_meta=False
#        )
#        self.repeat.write_info(
#            vr, key=INFOMETAS[self.annotdb_type]["repeat"]["ID"], add_meta=False
#        )
#
#        self.popfreq.write_info(
#            vr, key=INFOMETAS[self.annotdb_type]["popfreq"]["ID"], add_meta=False
#        )
#        self.cosmic.write_info(
#            vr, key=INFOMETAS[self.annotdb_type]["cosmic"]["ID"], add_meta=False
#        )
#
#    # POPFREQ ###################################################
#
#    def update_popfreq(self, vcfspec, dbsnp_vcf, search_equivs=True, overwrite=False):
#        assert (
#            self.annotdb_type == "plain"
#        ), f'This method is available only for annotdb_type "plain".'
#
#        try:
#            dbsnp_vr = fetch_dbsnp_vr(vcfspec, dbsnp_vcf, self.fasta, search_equivs)
#        except Exception as e:
#            raise Exception(f"dbsnp VCF fetch failed: " f"vcfspec: {vcfspec}") from e
#
#        self.popfreq.load_other(parse_dbsnp_vr(dbsnp_vr), overwrite=overwrite)
#
#    # COSMIC #####################################################
#
#    def update_cosmic(
#        self,
#        vcfspec,
#        cosmic_coding_vcf,
#        cosmic_noncoding_vcf,
#        search_equivs=True,
#        overwrite=False,
#    ):
#        assert (
#            self.annotdb_type == "plain"
#        ), f'This method is available only for annotdb_type "plain".'
#
#        cosmic_vr, cosmic_vr_noncoding = fetch_cosmic_vr(
#            vcfspec,
#            cosmic_coding_vcf,
#            cosmic_noncoding_vcf,
#            self.fasta,
#            search_equivs=search_equivs,
#        )
#        self.cosmic.load_other(
#            parse_cosmic_vr(cosmic_vr, cosmic_vr_noncoding), overwrite=overwrite
#        )
#
#    # FEATURES ###################################################
#
#    # post-cmdline-vep modifiers
#
#    def update_features_postvep_plain(
#        self,
#        vcfspec,
#        tabixfile_geneset,
#        tabixfile_regulatory,
#        tabixfile_repeats,
#        distance=veplib.DEFAULT_DISTANCE,
#    ):
#        """To be run with non-SV variants"""
#        assert self.annotdb_type == "plain"
#
#        mttype = vcfspec.get_mttype_firstalt()
#        assert mttype != "sv", f"Input variant must not be an SV."
#
#        # update using customfiles
#        fetch_interval = common.Interval(
#            chrom=vcfspec.chrom,
#            start1=max(1, vcfspec.pos - distance),
#            end1=min(self.chromdict[vcfspec.chrom], vcfspec.pos + distance),
#        )
#        self.update_customfile_transcript(
#            fetch_interval, tabixfile_geneset, fill_missing_is_canonical=True
#        )
#        self.update_customfile_regulatory(fetch_interval, tabixfile_regulatory)
#        self.update_customfile_repeat(fetch_interval, tabixfile_repeats)
#
#        # modifier
#        self.modifier_for_plain(vcfspec, mttype)
#        self.set_transcript_subsets()
#
#    def update_features_postvep_interval(
#        self,
#        interval,
#        tabixfile_geneset,
#        tabixfile_regulatory,
#        tabixfile_repeats,
#        distance=veplib.DEFAULT_DISTANCE,
#    ):
#        assert self.annotdb_type == "plain"
#
#        # update using customfiles
#        fetch_interval = common.Interval(
#            chrom=interval.chrom,
#            start1=max(1, interval.start1 - distance),
#            end1=min(self.chromdict[interval.chrom], interval.end1 + distance),
#        )
#        self.update_customfile_transcript(
#            fetch_interval, tabixfile_geneset, fill_missing_is_canonical=True
#        )
#        self.update_customfile_regulatory(fetch_interval, tabixfile_regulatory)
#        self.update_customfile_repeat(fetch_interval, tabixfile_repeats)
#
#        # modifier
#        self.modifier_for_interval(interval)
#        self.set_transcript_subsets()
#
#    def update_features_postvep_bnd(
#        self,
#        chrom,
#        pos,
#        endis5,
#        tabixfile_geneset,
#        tabixfile_regulatory,
#        tabixfile_repeats,
#        distance=veplib.DEFAULT_DISTANCE,
#    ):
#        assert self.annotdb_type in ("bnd1", "bnd2")
#
#        # update using customfiles
#        if endis5:
#            fetch_interval = common.Interval(
#                chrom=chrom, start1=pos, end1=min(self.chromdict[chrom], pos + distance)
#            )
#        else:
#            fetch_interval = common.Interval(
#                chrom=chrom, start1=max(1, pos - distance), end1=pos
#            )
#
#        self.update_customfile_transcript(
#            fetch_interval, tabixfile_geneset, fill_missing_is_canonical=True
#        )
#        self.update_customfile_regulatory(fetch_interval, tabixfile_regulatory)
#        self.update_customfile_repeat(fetch_interval, tabixfile_repeats)
#
#        # modifier
#        self.modifier_for_bnd(pos, endis5)
#        self.set_transcript_subsets()
#
#    # ensembl rest wrappers
#
#    def update_features_ensembl_wrapper_plain(
#        self,
#        vcfspec,
#        distance=veplib.DEFAULT_DISTANCE,
#        vep=False,
#        overlap=False,
#        transcript=False,
#        regulatory=False,
#    ):
#        """To be run with non-SV variants"""
#
#        assert self.annotdb_type == "plain"
#        mttype = vcfspec.get_mttype_firstalt()
#        assert mttype != "sv", f"Input variant must not be an SV."
#
#        if vep:
#            self.update_ensembl_rest_vep_plain(vcfspec, distance)
#        if overlap:
#            self.update_ensembl_rest_overlap_plain(vcfspec, distance)
#        if transcript:
#            self.update_ensembl_rest_lookup_transcript()
#        if regulatory:
#            self.update_ensembl_rest_regulatory()
#
#        self.modifier_for_plain(vcfspec, mttype)
#        self.set_transcript_subsets()
#
#    def update_features_ensembl_wrapper_interval(
#        self,
#        interval,
#        distance=veplib.DEFAULT_DISTANCE,
#        vep=False,
#        overlap=False,
#        transcript=False,
#        regulatory=False,
#    ):
#        assert self.annotdb_type == "plain"
#
#        if vep:
#            self.update_ensembl_rest_vep_interval(interval, distance)
#        if overlap:
#            self.update_ensembl_rest_overlap_interval(interval, distance)
#        if transcript:
#            self.update_ensembl_rest_lookup_transcript()
#        if regulatory:
#            self.update_ensembl_rest_regulatory()
#
#        self.modifier_for_interval(interval)
#        self.set_transcript_subsets()
#
#    def update_features_bnd(
#        self,
#        chrom,
#        pos,
#        endis5,
#        distance=veplib.DEFAULT_DISTANCE,
#        vep=False,
#        overlap=False,
#        transcript=False,
#        regulatory=False,
#    ):
#        assert self.annotdb_type in ("bnd1", "bnd2")
#
#        if vep:
#            self.update_ensembl_rest_vep_bnd(chrom, pos, distance)
#        if overlap:
#            self.update_ensembl_rest_overlap_bnd(chrom, pos, endis5, distance)
#        if transcript:
#            self.update_ensembl_rest_lookup_transcript()
#        if regulatory:
#            self.update_ensembl_rest_regulatory()
#
#        self.modifier_for_bnd(pos, endis5)
#        self.set_transcript_subsets()
#
#    # FEATURE UPDATE MODIFIERS ##########################################
#
#    # for setting missing distances
#
#    def set_missing_distances(self, distance_setter):
#        """Sets distance for annotitems not derived from VEP (e.g. repeat
#        elements). Simulates the distance given by VEP.
#        """
#
#        for attrname in ANNOTDB_ATTRNAMES_FEATURES:
#            annotitemdict = getattr(self, attrname)
#            for key, annotitem in annotitemdict.items():
#                if "distance" not in annotitem:
#                    # annotitems derived from "ensembl rest overlap"
#                    #   do not have "distance" value
#                    assert "start0" in annotitem and "end0" in annotitem, (
#                        f'"start0" and "end0" values are not set for this '
#                        f"AnnotItem object:\n{annotitem}"
#                    )
#                    feature_range0 = range(annotitem["start0"], annotitem["end0"])
#                    annotitem["distance"] = distance_setter(feature_range0)
#
#    def get_distance_setter_ins(self, vcfspec):
#        pos0 = vcfspec.pos - 1
#
#        def distance_setter(feature_range0):
#            if pos0 <= feature_range0.start - 1:
#                distance = feature_range0.start - pos0 - 1
#            elif pos0 >= feature_range0.stop - 1:
#                distance = feature_range0.stop - 1 - pos0
#            else:
#                distance = None
#
#            return distance
#
#        return distance_setter
#
#    def get_distance_setter_snv(self, vcfspec):
#        pos0 = vcfspec.pos - 1
#
#        def distance_setter(feature_range0):
#            if pos0 < feature_range0.start:
#                distance = feature_range0.start - pos0
#            elif pos0 >= feature_range0.stop:
#                distance = feature_range0.stop - 1 - pos0
#            else:
#                distance = None
#
#            return distance
#
#        return distance_setter
#
#    def get_distance_setter_interval(self, var_range0):
#        def distance_setter(feature_range0):
#            if var_range0.stop <= feature_range0.start:
#                distance = feature_range0.start - var_range0.stop + 1
#            elif var_range0.start >= feature_range0.stop:
#                distance = feature_range0.stop - 1 - var_range0.start
#            else:
#                distance = None
#
#            return distance
#
#        return distance_setter
#
#    def get_distance_setter_del(self, vcfspec):
#        var_range0 = range(vcfspec.pos, vcfspec.pos + len(vcfspec.ref) - 1)
#        return self.get_distance_setter_interval(var_range0)
#
#    def get_distance_setter_bnd(self, pos):
#        vcfspec = common.Vcfspec(None, pos, None, None)
#        return self.get_distance_setter_snv(vcfspec)
#
#    # modifier for plain
#    def modifier_for_plain(self, vcfspec, mttype):
#        if mttype == "snv":
#            distance_setter = self.get_distance_setter_snv(vcfspec)
#        elif mttype == "ins":
#            distance_setter = self.get_distance_setter_ins(vcfspec)
#        elif mttype in ("del", "delins", "mnv", "cpgmet"):
#            distance_setter = self.get_distance_setter_del(vcfspec)
#        else:
#            raise Exception(f"Unexpected mttype: {mttype}")
#
#        self.set_missing_distances(distance_setter)
#
#    # modifier for interval
#    def modifier_for_interval(self, interval):
#        distance_setter = self.get_distance_setter_interval(interval.range0)
#        self.set_missing_distances(distance_setter)
#        self.set_is_enclosed(interval.range0)
#
#    def set_is_enclosed(self, var_range0):
#        """
#        Sets "is_enclosed" attribute. Must be run after "start0" and "end0"
#        values have been set for all annotitems.
#        """
#
#        for attrname in ANNOTDB_ATTRNAMES_FEATURES:
#            annotitemdict = getattr(self, attrname)
#            for key, annotitem in annotitemdict.items():
#                feature_range0 = range(annotitem["start0"], annotitem["end0"])
#                annotitem["is_enclosed"] = (
#                    feature_range0.start >= var_range0.start
#                    and feature_range0.stop <= var_range0.stop
#                )
#
#    # modifier for breakend
#    def modifier_for_bnd(self, pos, endis5):
#        distance_setter = self.get_distance_setter_bnd(pos)
#        self.set_missing_distances(distance_setter)
#
#        self.settle_border_issues(pos, endis5)
#
#    def settle_border_issues(self, pos, endis5):
#        """
#        Must be run after "set_missing_distances" and "lookup_transcript" or
#        "lookup_regulatory" has been run.
#
#        1) Removes features entirely on the other side
#        2) For features whose border coincide with the breakend border,
#        distance is changed from None to 0.
#        3) Sets "is_broken" attribute
#        """
#
#        (on_the_other_side, on_the_border) = self.get_feature_classifiers(endis5)
#
#        for attrname in ANNOTDB_ATTRNAMES_FEATURES:
#            annotitemdict = getattr(self, attrname)
#            # step 1) and 2)
#            keys_to_delete = list()
#            for key, annotitem in annotitemdict.items():
#                if on_the_other_side(annotitem):
#                    keys_to_delete.append(key)
#                else:
#                    if annotitem["distance"] is None:
#                        if on_the_border(annotitem, pos):
#                            annotitem["distance"] = 0
#            for key in keys_to_delete:
#                del annotitemdict[key]
#            # step 3)
#            for key, annotitem in annotitemdict.items():
#                annotitem["is_broken"] = annotitem["distance"] is None
#
#    def get_feature_classifiers(self, endis5):
#        """
#                (endis5 == True)
#                                   THE OTHER SIDE    THE BREAKEND SIDE
#                                   - - - - - - - -  -----------------
#        FEATURE ON THE OTHER SIDE  = = = = = =     | = = = = = = = =  : FEATURE ON THE BORDER
#                                   - - - - - - - -  -----------------
#                                   A C G T A C G T   A C G T A C G T
#                                                     *(pos)
#                ("= = =" denotes a feature)
#        """
#
#        if endis5:
#
#            def on_the_other_side(annotitem):
#                return (annotitem["distance"] is not None) and (
#                    annotitem["distance"] < 0
#                )
#
#            def on_the_border(annotitem, pos):
#                return annotitem["start1"] == pos
#
#        else:
#
#            def on_the_other_side(annotitem):
#                return (annotitem["distance"] is not None) and (
#                    annotitem["distance"] > 0
#                )
#
#            def on_the_border(annotitem, pos):
#                return annotitem["end1"] == pos
#
#        return on_the_other_side, on_the_border
#
#    # CUSTOM FILE FETCH FUNCTIONS #####################################
#
#    def update_customfile_repeat(self, fetch_interval, tabixfile_repeats):
#        repeat_dict = customfile.fetch_repeat(
#            fetch_interval.chrom,
#            fetch_interval.start0,
#            fetch_interval.end0,
#            tabixfile_repeats,
#        )
#        self.repeat.load_other(repeat_dict, overwrite=False, create_new=True)
#
#    def update_customfile_transcript(
#        self, fetch_interval, tabixfile_geneset, fill_missing_is_canonical=True
#    ):
#        transcript_dict = customfile.fetch_transcript(
#            fetch_interval.chrom,
#            fetch_interval.start0,
#            fetch_interval.end0,
#            tabixfile_geneset,
#        )
#        self.transcript.load_other(transcript_dict, overwrite=False, create_new=True)
#        if fill_missing_is_canonical:
#            for annotitem_id, annotitem in self.transcript.items():
#                if "is_canonical" not in annotitem:
#                    raw_result = ensembl_rest.lookup_id(
#                        ID=annotitem["id"], refver=self.refver, expand=False
#                    )
#                    if raw_result["is_canonical"] == 1:
#                        annotitem["is_canonical"] = True
#                    else:
#                        annotitem["is_canonical"] = False
#
#    def update_customfile_regulatory(self, fetch_interval, tabixfile_regulatory):
#        regulatory_dict = customfile.fetch_regulatory(
#            fetch_interval.chrom,
#            fetch_interval.start0,
#            fetch_interval.end0,
#            tabixfile_regulatory,
#        )
#        self.regulatory.load_other(regulatory_dict, overwrite=False, create_new=True)
#
#    # ENSEMBL UNIT FUNCTIONS ##########################################
#
#    # vep
#    def update_ensembl_rest_vep_plain(self, vcfspec, distance=veplib.DEFAULT_DISTANCE):
#        self._update_ensembl_rest_vep_helper(
#            vcfspec=vcfspec, hgvsg=None, distance=distance
#        )
#
#    def update_ensembl_rest_vep_interval(
#        self, interval, distance=veplib.DEFAULT_DISTANCE
#    ):
#        sv_del = structvars.Deletion(
#            interval.chrom, interval.start1, interval.end1, fasta=self.fasta
#        )
#        hgvsg = sv_del.get_hgvsg()
#        self._update_ensembl_rest_vep_helper(
#            vcfspec=None, hgvsg=hgvsg, distance=distance
#        )
#
#    def update_ensembl_rest_vep_bnd(self, chrom, pos, distance=veplib.DEFAULT_DISTANCE):
#        ref = self.fasta.fetch(chrom, pos - 1, pos)
#        vcfspec = common.Vcfspec(chrom, pos, ref, ["N"])
#        self._update_ensembl_rest_vep_helper(
#            vcfspec=vcfspec, hgvsg=None, distance=distance
#        )
#
#    # overlap
#    def update_ensembl_rest_overlap_plain(
#        self, vcfspec, distance=veplib.DEFAULT_DISTANCE
#    ):
#        chrom = vcfspec.chrom
#        start1 = vcfspec.pos - distance
#        end1 = vcfspec.pos + distance
#        interval = common.Interval(chrom, start1, end1)
#        self._update_ensembl_rest_overlap_helper(interval)
#
#    def update_ensembl_rest_overlap_interval(
#        self, interval, distance=veplib.DEFAULT_DISTANCE
#    ):
#        new_interval = common.Interval(
#            chrom=interval.chrom,
#            start1=max(1, interval.start1 - distance),
#            end1=min(self.chromdict[interval.chrom], interval.end1 + distance),
#        )
#        self._update_ensembl_rest_overlap_helper(new_interval)
#
#    def update_ensembl_rest_overlap_bnd(
#        self, chrom, pos, endis5, distance=veplib.DEFAULT_DISTANCE
#    ):
#        if endis5:
#            start1 = pos
#            end1 = min(self.chromdict[chrom], pos + distance)
#        else:
#            start1 = max(1, pos - distance)
#            end1 = pos
#        interval = common.Interval(chrom, start1, end1)
#        self._update_ensembl_rest_overlap_helper(interval)
#
#    # lookup_transcript
#    def update_ensembl_rest_lookup_transcript(self):
#        parsed = ensembl_parser.parse_rest_lookup_transcript_post(
#            ensembl_rest.lookup_id_post(
#                tuple(self.transcript.keys()), refver=self.refver, expand=True
#            ),
#            refver=self.refver,
#            set_gene_name=True,
#        )
#        self._load_ensembl_parsed(parsed, overwrite=False, create_new=False)
#
#    # regulatory
#    def update_ensembl_rest_regulatory(self):
#        for ID in self.regulatory.keys():
#            self._update_ensembl_rest_regulatory_helper(ID)
#
#    # cmdline VEP
#    def update_cmdline_vep(self, vr, overwrite=True, create_new=True):
#        try:
#            parsed = ensembl_parser.parse_cmdline_vep(vr)
#        except Exception as e:
#            raise Exception(
#                f"Failure of VEP output variant record parsing:" f"\n{vr}"
#            ) from e
#
#        self._load_ensembl_parsed(parsed, overwrite=overwrite, create_new=create_new)
#
#    # helpers
#    def _load_ensembl_parsed(self, parsed, overwrite=False, create_new=False):
#        for annotdb_attrname, other_annotitemdict in parsed.items():
#            annotitemdict = getattr(self, annotdb_attrname)
#            annotitemdict.load_other(
#                other_annotitemdict, overwrite=overwrite, create_new=create_new
#            )
#
#    def _update_ensembl_rest_vep_helper(self, vcfspec, hgvsg, distance):
#        parsed = ensembl_parser.parse_rest_vep(
#            ensembl_rest.vep(
#                refver=self.refver,
#                vcfspec=vcfspec,
#                hgvsg=hgvsg,
#                distance=distance,
#                with_CADD=True,
#                with_Phenotypes=False,
#                with_canonical=True,
#                with_mane=True,
#                with_miRNA=False,
#                with_numbers=True,
#                with_protein=True,
#                with_ccds=True,
#                with_hgvs=True,
#            )
#        )
#        self._load_ensembl_parsed(parsed, overwrite=False, create_new=True)
#
#    def _update_ensembl_rest_overlap_helper(self, interval):
#        parsed = ensembl_parser.parse_rest_overlap(
#            ensembl_rest.overlap(
#                chrom=interval.chrom,
#                start1=interval.start1,
#                end1=interval.end1,
#                refver=self.refver,
#                transcript=False,
#                regulatory=False,
#                motif=False,
#                repeat=True,
#            ),
#            refver=self.refver,
#            include_motif_without_evidence=False,
#        )
#        self._load_ensembl_parsed(parsed, overwrite=False, create_new=True)
#
#    def _update_ensembl_rest_lookup_transcript_helper(self, ID):
#        parsed = ensembl_parser.parse_rest_lookup_transcript(
#            ensembl_rest.lookup_id(ID, refver=self.refver, expand=True),
#            refver=self.refver,
#            set_gene_name=True,
#        )
#        self._load_ensembl_parsed(parsed, overwrite=False, create_new=False)
#
#    def _update_ensembl_rest_regulatory_helper(self, ID):
#        parsed = ensembl_parser.parse_rest_regulatory(
#            ensembl_rest.regulatory(ID, refver=self.refver)
#        )
#        self._load_ensembl_parsed(parsed, overwrite=False, create_new=False)
#
#    # FEATURE ADDITIONAL ATTRIBUTES GENERATION #####################
#
#    def set_transcript_subsets(self):
#        # canonical, overlap, both
#        self.transcript_canon = AnnotItemDict()
#        self.transcript_ovlp = AnnotItemDict()
#        self.transcript_canon_ovlp = AnnotItemDict()
#
#        for ID, annotitem in self.transcript.items():
#            if annotitem["is_canonical"]:
#                self.transcript_canon[ID] = annotitem
#            if annotitem["distance"] is None:
#                self.transcript_ovlp[ID] = annotitem
#            if annotitem["is_canonical"] and (annotitem["distance"] is None):
#                self.transcript_canon_ovlp[ID] = annotitem
#
#    def set_exon_intron_ranges(self):
#        def make_exon_ranges(annotitem):
#            exon_ranges = dict()
#            for key, val in annotitem["exons"].items():
#                keysp = key.split("_")
#                valsp = [int(x) for x in val.split("_")]
#                exon_number = int(keysp[0])
#                exon_ranges[exon_number] = range(valsp[0], valsp[1])
#            return exon_ranges
#
#        def make_intron_ranges(exon_ranges, annotitem):
#            if len(exon_ranges) == 1:
#                intron_ranges = None
#            else:
#                intron_ranges = dict()
#                ascending = sorted(exon_ranges.items(), key=lambda x: x[1][0])
#                for idx in range(len(ascending) - 1):
#                    intron_number = (
#                        ascending[idx][0]
#                        if annotitem["is_forward"]
#                        else ascending[idx + 1][0]
#                    )
#                    range_item = range(
#                        ascending[idx][1][-1] + 1, ascending[idx + 1][1][0]
#                    )
#                    intron_ranges[intron_number] = range_item
#                intron_ranges = dict(sorted(intron_ranges.items(), key=lambda x: x[0]))
#
#            return intron_ranges
#
#        for ID, annotitem in self.transcript.items():
#            if "exons" in annotitem.keys():
#                annotitem.exon_ranges = make_exon_ranges(annotitem)
#                annotitem.intron_ranges = make_intron_ranges(
#                    annotitem.exon_ranges, annotitem
#                )

##############################
# Base classes - AnnotItemBase
##############################

class AllBase:
    def __repr__(self):
        if self.is_missing:
            return (
                f'<{self.__class__.__name__}(MISSING)>'
            )
        else:
            return (
                f'<{self.__class__.__name__}(\n'
                f'{textwrap.indent(tools.cpformat(self.get_self_show(), sort_dicts=False), " " * 4)}'
                f'\n)>'
            )

#    @property
#    def annotkey(self):
#        if hasattr(self.__class__, 'meta'):
#            return self.__class__.meta["ID"]
#        else:
#            return self.__class__.unit_class["ID"]

        #return self.__class__.get_meta()["ID"]

    @staticmethod
    def decode(infostring):
        return _decode(infostring)

    @classmethod
    def get_annotkey(cls):
        return cls.meta['ID']
        #if hasattr(cls, 'meta'):
        #    return cls.meta["ID"]
        #else:
        #    return cls.unit_class["ID"]

#    @classmethod
#    def get_meta(cls):
#        if hasattr(cls, 'meta'):
#            return getattr(cls, 'meta')
#        else:
#            return getattr(cls.unit_class, 'meta')


class InfoBase:
    @classmethod
    def add_meta(cls, vcfheader):
        vcfheader.add_meta(key="INFO", items=cls.meta.items())
        #if hasattr(cls, 'meta'):
        #    vcfheader.add_meta(key="INFO", items=cls.meta.items())
        #else:
        #    vcfheader.add_meta(key="INFO", items=cls.unit_class.meta.items())

    def write_base(self, vr):
        if self.is_missing:
            raise Exception(f'A missing AnnotItem object cannot be written.')

        annotkey = self.get_annotkey()
        if annotkey not in vr.header.info.keys():
            self.add_meta(vr.header)
        infoformat.set_info(vr, annotkey, self.encode())

    @staticmethod
    def get_Number(cls, vr, annotkey):
        if annotkey in vr.header.info.keys():
            result = infoformat.get_translated_number_info(vr, annotkey)
            if isinstance(result, int):
                return result
            else:
                return None
        else:
            return None


class FormatBase:
    @classmethod
    def add_meta(cls, vcfheader):
        vcfheader.add_meta(key="FORMAT", items=cls.meta.items())
        #if hasattr(cls, 'meta'):
        #    vcfheader.add_meta(key="FORMAT", items=cls.meta.items())
        #else:
        #    vcfheader.add_meta(key="FORMAT", items=cls.unit_class.meta.items())

#    def write_base(self, vr, sampleid):
#        self.meta_sanitycheck()
#
#        annotkey = self.get_annotkey()
#        if annotkey not in vr.header.formats.keys():
#            self.add_meta(vr.header)
#        infoformat.set_format(vr, sampleid, annotkey, self.encode())
#
#    def write_base_old(self, vr, sampleid, annotkey):
#        infoformat.set_format(vr, sampleid, annotkey, self.encode())

    @staticmethod
    def get_Number(vr, annotkey):
        if annotkey in vr.header.formats.keys():
            result = infoformat.get_translated_number_format(vr, annotkey)
            if isinstance(result, int):
                return result
            else:
                return None
        else:
            return None


##############################



############################################################################
# Base classes - SingleItemBase, ALTlistBase #
############################################################################

class SingleItemBase(AllBase, dict):
    ##### constructors #####
    def __init__(self, is_missing):
        self.is_missing = is_missing

    @classmethod
    def from_raw_value(cls, vr_value):
        if infoformat.check_InfoFormatValue_isNA(vr_value):
            return cls(is_missing=True)
        else:
            result = cls(is_missing=False)
            result.update(cls.decode(vr_value))
            return result

    from_annotstring = from_raw_value
    from_infostring = from_annotstring

    @classmethod
    def from_dict(cls, data):
        return cls(data)

    @classmethod
    def init_missing(cls):
        return cls(is_missing=True)

    @classmethod
    def init_nonmissing(cls):
        return cls(is_missing=False)
    ########################

    def meta_sanitycheck(self):
        if self.__class__.meta['Number'] != 1:
            raise Exception(f'"Number" value of metadata must be 1 for a subclass of SingleItemBase class.')

    def get_self_show(self):
        return dict(self)

    #def repr_base(self):
    #    return tools.cpformat(dict(self), sort_dicts=False)

    def set_is_missing(self):
        self.is_missing = True

    def set_is_not_missing(self):
        self.is_missing = False

    def encode(self):
        if self.is_missing:
            return None
        else:
            return _encode(dict(self), handle_percent=True)

    def update_vr_value(self, vr_value, overwrite=True):
        self.update_dict(self.__class__.decode(vr_value), overwrite=overwrite)

    update_annotstring = update_vr_value
    update_infostring = update_annotstring
    load_infostring = update_annotstring

    def update_dict(self, other, overwrite=False):
        if overwrite:
            for key, val in other.items():
                self[key] = val
        else:
            for key, new_val in other.items():
                try:
                    old_val = self[key]
                except KeyError:
                    self[key] = new_val
                else:
                    if old_val is None:
                        self[key] = new_val
                    else:
                        if new_val != old_val:
                            self._update_conflict_handler(key, old_val, new_val)

    load_other = update_dict

    def _update_conflict_handler(self, key, old_val, new_val):
        values = {'3prime_overlapping_ncRNA', '3prime_overlapping_ncrna'}
        if (
            key == 'biotype' and 
            old_val in values and 
            new_val in values
        ):
            self[key] = '3prime_overlapping_ncRNA'
        else:
            raise Exception(f"Old value and new value are different; key: {key}; old_val: {old_val}; new_val: {new_val}")


class ALTlistBase(AllBase, list):
    @classmethod
    def from_raw_value(cls, vr_value):
        assert isinstance(vr_value, tuple)
        result = cls()
        for x in vr_value:
            unit = cls.unit_class.from_raw_value(x)
            result.append(unit)

        return result

    from_annotstring_tuple = from_raw_value

    @classmethod
    def init_missing(cls, vr):
        result = cls()
        for _ in range(len(vr.alts)):
            result.append(cls.unit_class(is_missing=True))

        return result

#    def repr_base(self):
#        return pprint.pformat([x.repr_base() for x in self])
#        #return '[' + '\n'.join(x.repr_base() for x in self) + ']'
#        #return tuple(x.get_self_show() for x in self)

    def meta_sanitycheck(self):
        if self.__class__.meta['Number'] != 'A':
            raise Exception(f'"Number" value of metadata must be "A" for a subclass of "ALTlistBase" class.')

    def get_self_show(self):
        return list(self)

    def encode(self):
        return tuple(x.encode() for x in self)

    @property
    def is_missing(self):
        return all(x.is_missing for x in self)

    #def append(self, x):
    #    assert isinstance(x, self.__class__.unit_class), f'Only an instance of unit_class can be appended.'
    #    super().append(x)
#
#    def extend(self, seq):
#        assert all(isinstance(x, self.__class__.unit_class) for x in seq), f'Only an instance of unit_class can be appended.'
#        super().extend(seq)

############################################################################


######################################################
# Intermediate level base classes for FORMAT entries #
######################################################

class AnnotItemFormatSingle(SingleItemBase, FormatBase):
    pass

AnnotItemVariantFormatSingle = AnnotItemFormatSingle


class AnnotItemFormatALTList(ALTlistBase, FormatBase):
    pass

AnnotItemVariantFormatALTList = AnnotItemFormatALTList


################################################################################
# Objects which can be read from and written to VariantRecord or VariantHeader #
################################################################################

class AnnotItemHeader(SingleItemBase):
    @classmethod
    def from_vcfheader_base(cls, vcfheader):
        """Returns:
            None if the vcfheader does not contain annotkey
        """
        result = None
        annotkey = cls.get_annotkey()
        for rec in vcfheader.records:
            if rec.key == annotkey:
                result = cls.from_raw_value(rec.value)

        return result

    def write_base(self, vcfheader):
        if self.is_missing:
            raise Exception(f'A missing AnnotItem object cannot be written.')

        vcfheader.add_meta(key=self.get_annotkey(), value=self.encode())


class AnnotItemInfoSingle(SingleItemBase, InfoBase):
    @classmethod
    def from_vr_base(cls, vr, *args, **kwargs):
        annotkey = cls.get_annotkey()
        if infoformat.check_NA_info(vr, annotkey):
            return cls.init_missing()
        else:
            raw_val = vr.info[annotkey]
            assert not isinstance(raw_val, tuple)
            return cls.from_raw_value(raw_val, *args, **kwargs)


AnnotItemVariantInfoSingle = AnnotItemInfoSingle


class AnnotItemInfoALTlist(ALTlistBase, InfoBase):
    @classmethod
    def from_vr_base(cls, vr, *args, **kwargs):
        annotkey = cls.get_annotkey()
        if infoformat.check_NA_info(vr, annotkey):
            return cls.init_missing(vr)
        else:
            raw_val = vr.info[annotkey]
            assert isinstance(raw_val, tuple)
            return cls.from_raw_value(raw_val, *args, **kwargs)


AnnotItemVariantInfoALTlist = AnnotItemInfoALTlist


class AnnotItemFormatSampledict(AllBase, FormatBase, dict):
    #@classmethod
    #def get_annotkey(cls):
    #    return cls.unit_class.get_annotkey()

    #@property
    #def annotkey(self):
    #    return self.__class__.unit_class.get_annotkey()

    @classmethod
    def from_vr_base(cls, vr, sampleid_list=None):
        if sampleid_list is None:
            sampleid_list = list(vr.samples.keys())
        else:
            sampleid_list = np.atleast_1d(sampleid_list)

        result = cls()
        annotkey = cls.get_annotkey()
        for sampleid in sampleid_list:
            if infoformat.check_NA_format(vr, sampleid, annotkey):
                result[sampleid] = cls.unit_class.init_missing()
            else:
                result[sampleid] = cls.unit_class.from_raw_value(
                    vr.samples[sampleid][annotkey]
                )

        return result

    @classmethod
    def init_missing(cls, vr):
        result = cls()
        for sampleid in vr.samples.keys():
            result[sampleid] = cls.unit_class.init_missing()

        return result

    def get_self_show(self):
        return dict(self)

    @property
    def is_missing(self):
        return all(x.is_missing for x in self.values())

    def get_sample_augmented_vr(self, vr):
        """When some of sample IDs of self are missing from input vr, make
        a copy of input vr with the missing sample IDs added. Otherwise,
        returns the input vr.
        """
        import handygenome.variant.varianthandler as varianthandler

        self_samples = set(self.keys())
        vr_samples = set(vr.samples.keys())
        diff_samples = self_samples.difference(vr_samples)
        if len(diff_samples) > 0:
            new_header = vr.header.copy()
            for sampleid in diff_samples:
                new_header.add_sample(sampleid)
            return varianthandler.reheader(vr, new_header)
        else:
            return vr

    def write_base(self, vr, donot_write_missing=True):
        if self.is_missing:
            raise Exception(f'A missing AnnotItem object cannot be written.')

        edited_vr = self.get_sample_augmented_vr(vr)

        annotkey = self.get_annotkey()
        if annotkey not in edited_vr.header.formats.keys():
            self.add_meta(edited_vr.header)

        for sampleid, annotitem in self.items():
            if donot_write_missing:
                if not annotitem.is_missing:
                    infoformat.set_format(edited_vr, sampleid, annotkey, annotitem.encode())
            else:
                infoformat.set_format(edited_vr, sampleid, annotkey, annotitem.encode())

        return edited_vr


##########################


#class AnnotItem(AnnotItemBase, dict):
#    meta = None
#    # to be used as "items" argument for pysam.VariantHeader.add_meta
#
#    ##### constructors #####
#    @classmethod
#    def from_infostring(cls, infostring):
#        result = cls()
#        result.update(cls.decode(infostring))
#        return result
#
#    @classmethod
#    def from_vcfheader(cls, vcfheader):
#        result = None
#        annotkey = cls.get_annotkey()
#        for rec in vcfheader.records:
#            if rec.key == annotkey:
#                result = cls.from_infostring(rec.value)
#        if result is None:
#            raise Exception(
#                f"The vcfheader does not contain annotation key {annotkey}."
#            )
#
#        return result
#
#    @classmethod
#    def from_dict(cls, data):
#        return cls(data)
#    ########################
#
#    def encode(self):
#        return _encode(dict(self))
#
#    def load_infostring(self, infostring, overwrite=True):
#        self.update_dict(self.__class__.decode(infostring), overwrite=overwrite)
#    
#    #    def load_vr_info(self, vr, infokey=None):
#    #        if infokey is None:
#    #            infokey = self.annotkey
#    #
#    #        assert vr.header.info[infokey].number == 1
#    #        assert vr.header.info[infokey].type == "String"
#    #
#    #        if not infoformat.check_NA_info(vr, infokey):
#    #            infostring = infoformat.get_info(vr, infokey)
#    #            self.load_infostring(infostring)
#    #
#    #    load_vr = load_vr_info  # alias
#    #
#    #    def load_vr_format(self, vr, sampleid, key=None):
#    #        if key is None:
#    #            key = self.annotkey
#    #
#    #        assert vr.header.formats[key].number == 1
#    #        assert vr.header.formats[key].type == "String"
#    #
#    #        if not infoformat.check_NA_format(vr, sampleid, key):
#    #            infostring = infoformat.get_format(vr, sampleid, key)
#    #            self.load_infostring(infostring)
#
#    def update_dict(self, other, overwrite=False):
#        if overwrite:
#            for key, val in other.items():
#                self[key] = val
#        else:
#            for key, new_val in other.items():
#                try:
#                    old_val = self[key]
#                except KeyError:
#                    self[key] = new_val
#                else:
#                    if old_val is None:
#                        self[key] = new_val
#                    else:
#                        if new_val != old_val:
#                            self._update_conflict_handler(key, old_val, new_val)
#
#    def _update_conflict_handler(self, key, old_val, new_val):
#        values = {'3prime_overlapping_ncRNA', '3prime_overlapping_ncrna'}
#        if (
#            key == 'biotype' and 
#            old_val in values and 
#            new_val in values
#        ):
#            self[key] = '3prime_overlapping_ncRNA'
#        else:
#            raise Exception(f"Old value and new value are different; key: {key}; old_val: {old_val}; new_val: {new_val}")
#
#
#    load_other = update_dict
#
#
#    def write_info(self, vr, key=None, add_meta=False):
#        if key is None:
#            key = self.annotkey
#        if add_meta:
#            self.__class__.add_meta_info(vr.header)
#
#        infoformat.set_info(vr, key, self.encode())
#
#    # write = write_info  # alias
#
#    def write_format(self, vr, sampleid, key=None, add_meta=False):
#        if key is None:
#            key = self.annotkey
#        if add_meta:
#            self.__class__.add_meta_format(vr.header)
#
#        infoformat.set_format(vr, sampleid, key, self.encode())




#class AnnotItemDict(dict):
#    def __init__(self, vr=None, infokey=None, septype=DEFAULT_SEPTYPE):
#        # super().__init__()
#        self.septype = septype
#        if (vr is not None) and (infokey is not None):
#            self.load_vr(vr, infokey)
#
#    def __setitem__(self, key, val):
#        assert isinstance(val, AnnotItem)
#        super().__setitem__(key, val)
#
#    def __repr__(self):
#        return tools.cpformat(self.get_showdict())
#
#    def get_showdict(self):
#        showdict = {key: dict(val) for (key, val) in self.items()}
#        return showdict
#
#    def get_annotitem_id(self, annotitem):
#        pass
#
#    def load_vr(self, vr, infokey):
#        assert vr.header.info[infokey].number == "."
#        assert vr.header.info[infokey].type == "String"
#
#        if not infoformat.check_NA_info(vr, infokey):
#            infoval = infoformat.get_info(vr, infokey, collapse_tuple=False)
#            for infostring in infoval:
#                annotitem = AnnotItem(septype=self.septype)
#                annotitem.load_infostring(infostring)
#                ID = self.get_annotitem_id(annotitem)
#                self[ID] = annotitem
#
#    def load_other(self, other, overwrite=True, create_new=False):
#        for ID, annotitem in other.items():
#            if ID in self.keys():
#                self[ID].load_other(annotitem, overwrite=overwrite)
#            else:
#                if create_new:
#                    self[ID] = annotitem
#
#    def write_info(self, vr, key, add_meta=False):
#        if add_meta:
#            add_infometas(vr.header)
#
#        infoval = list()
#        for annotitem in self.values():
#            infoval.append(annotitem.encode())
#
#        vr.info[key] = infoval



import collections

import pysam

import importlib
top_package_name = __name__.split('.')[0]
common = importlib.import_module('.'.join([top_package_name, 'common']))
readhandler = importlib.import_module('.'.join([top_package_name, 'readplus', 'readhandler']))


def call_pileup(bam, chrom, start0, end0):
    pup = bam.pileup(chrom, start0, end0, 
                     truncate=True, max_depth=10**6, stepper='all', 
                     ignore_overlaps=False, ignore_orphans=False, 
                     min_base_quality=0, min_mapping_quality=0, 
                     adjust_capq_threshold=0, compute_baq=False, 
                     redo_baq=False)    

    return pup

    
def get_max_vaf_pysampileup(pupcol):
    counts = collections.Counter(pupcol)
    max_vaf = max(counts.values()) / sum(counts.values())
    return max_vaf


def get_is_active_pupcol(pupcol, threshold):
    max_vaf = get_max_vaf_pysampileup(
        x.upper() for x in
        pupcol.get_query_sequences(mark_matches=False, mark_ends=False, add_indels=True)
    )
    is_active = max_vaf <= threshold
    return pupcol.reference_pos, is_active


def get_active_info_from_pileup(pup, threshold):
    return [get_is_active_pupcol(pupcol, threshold) for pupcol in pup]


def check_active_col_pysampileup_rightward(bam, chrom, pos0_init, threshold, add_pileup_by):
    start0 = pos0_init
    end0 = start0 + add_pileup_by
    while True:
        pup = call_pileup(bam, chrom, start0, end0)
        for pupcol in pup:
            pos0, is_active = get_is_active_pupcol(pupcol, threshold)
            yield pos0, is_active

        start0 = end0
        end0 = start0 + add_pileup_by
        continue

        
def check_active_col_pysampileup_leftward(bam, chrom, pos0_init, threshold, add_pileup_by):
    end0 = pos0_init + 1
    start0 = end0 - add_pileup_by
    while True:
        pup = call_pileup(bam, chrom, start0, end0)
        active_info_buffer = get_active_info_from_pileup(pup, threshold)
        for pos0, is_active in reversed(active_info_buffer):
            yield pos0, is_active

        end0 = start0
        start0 = end0 - add_pileup_by
        continue

            
def get_active_range_pysampileup(bam, chrom, start0, end0, threshold=0.9, inactive_padding=5, add_pileup_by=10):
#     readlist = list(bam.fetch(chrom, start0, end0))
#     pileup_df = get_pileup(readlist, as_df=True)
    pup = call_pileup(bam, chrom, start0, end0)
    naive_active_info = get_active_info_from_pileup(pup, threshold)        
    naive_active_positions = [pos0 for (pos0, is_active) in naive_active_info if is_active]    
    print(naive_active_info)
    if len(naive_active_positions) == 0:
        return
    
    naive_active_positions_min = min(naive_active_positions)
    naive_active_positions_max = max(naive_active_positions)
    
#     idx_min = naive_active_info.index((naive_active_positions_min, True))
#     idx_max = naive_active_info.index((naive_active_positions_max, True))
#     naive_active_info_filtered = naive_active_info[idx_min:(idx_min + 1)]
    
    active_info_rightward = list()
    active_info_leftward = list()
    
    # rightward
    current_pos0 = naive_active_positions_max + 1
    for pos0, is_active in check_active_col_pysampileup_rightward(
        bam, chrom, current_pos0, threshold=threshold, add_pileup_by=add_pileup_by
    ):
        active_info_rightward.append((pos0, is_active))
        if len(active_info_rightward) >= inactive_padding:
            if not any(x[1] for x in active_info_rightward[-inactive_padding:]):
                break

    # leftward
    current_pos0 = naive_active_positions_min - 1
    for pos0, is_active in check_active_col_pysampileup_leftward(
        bam, chrom, current_pos0, threshold=threshold, add_pileup_by=add_pileup_by
    ):
        active_info_leftward.append((pos0, is_active))
        if len(active_info_leftward) >= inactive_padding:
            if not any(x[1] for x in active_info_leftward[-inactive_padding:]):
                break
                
#     active_info = active_info_leftward[::-1] + naive_active_info_filtered + active_info_rightward
#     positions = [pos0 for (pos0, is_active) in active_info]    
    active_range = range(active_info_leftward[-1][0], active_info_rightward[-1][0] + 1)
    
    return active_range


###############################

def get_max_vaf_mypileup(col):
    counts = collections.Counter(x for x in col if x is not None)
    max_vaf = max(counts.values()) / sum(counts.values())    
    return max_vaf

def get_iterator(pileup, as_array, start0, end0):
    if as_array:
        rng = range(start0, end0)
        return ((rng[idx], pileup[:, idx]) for idx in range(pileup.shape[1]))
    else:
        return pileup.items()

    
def get_active_info_mypileup(bam, chrom, start0, end0, threshold, as_array):
    pileup = readhandler.get_pileup(bam, chrom, start0, end0, as_array=as_array, truncate=True)
    result = list()

    for pos0, col in get_iterator(pileup, as_array, start0, end0):
        is_active = (get_max_vaf_mypileup(col) <= threshold)
        result.append((pos0, is_active))
    return result


def check_active_col_mypileup_rightward(bam, chrom, pos0_init, threshold, add_pileup_by, as_array):
    start0 = pos0_init
    end0 = start0 + add_pileup_by
    while True:
        pileup = readhandler.get_pileup(bam, chrom, start0, end0, as_array=as_array, truncate=True)
        for pos0, col in get_iterator(pileup, as_array, start0, end0):
            is_active = (get_max_vaf_mypileup(col) <= threshold)
            yield pos0, is_active

        start0 = end0
        end0 = start0 + add_pileup_by
        continue
        
        
def check_active_col_mypileup_leftward(bam, chrom, pos0_init, threshold, add_pileup_by, as_array):
    end0 = pos0_init + 1
    start0 = end0 - add_pileup_by
    while True:
        pileup = readhandler.get_pileup(bam, chrom, start0, end0, as_array=as_array, truncate=True)
        for pos0, col in reversed(tuple(get_iterator(pileup, as_array, start0, end0))):
            is_active = (get_max_vaf_mypileup(col) <= threshold)
            yield pos0, is_active

        end0 = start0
        start0 = end0 - add_pileup_by
        continue    
    
###
    

def get_active_range_mypileup(bam, chrom, start0, end0, threshold=0.9, inactive_padding=5, add_pileup_by=10, as_array=False):
    naive_active_info = get_active_info_mypileup(bam, chrom, start0, end0, threshold, as_array)
    naive_active_positions = [pos0 for (pos0, is_active) in naive_active_info if is_active]    
    print(naive_active_info)
    if len(naive_active_positions) == 0:
        return
    
    naive_active_positions_min = min(naive_active_positions)
    naive_active_positions_max = max(naive_active_positions)
        
    active_info_rightward = list()
    active_info_leftward = list()
    
    # rightward
    current_pos0 = naive_active_positions_max + 1
#     n = 0
    for pos0, is_active in check_active_col_mypileup_rightward(
        bam, chrom, current_pos0, threshold=threshold, add_pileup_by=add_pileup_by, as_array=as_array,
    ):
#         n += 1
#         if n == 500:
#             print('overflow')
#             break
        active_info_rightward.append((pos0, is_active))
        if len(active_info_rightward) >= inactive_padding:
            if not any(x[1] for x in active_info_rightward[-inactive_padding:]):
                break

    # leftward
    current_pos0 = naive_active_positions_min - 1
#     n = 0
    for pos0, is_active in check_active_col_mypileup_leftward(
        bam, chrom, current_pos0, threshold=threshold, add_pileup_by=add_pileup_by, as_array=as_array,
    ):
#         n += 1
#         if n == 500:
#             print('overflow')
#             break
        active_info_leftward.append((pos0, is_active))
        if len(active_info_leftward) >= inactive_padding:
            if not any(x[1] for x in active_info_leftward[-inactive_padding:]):
                break
                
#     active_info = active_info_leftward[::-1] + naive_active_info_filtered + active_info_rightward
#     positions = [pos0 for (pos0, is_active) in active_info]    
    active_range = range(active_info_leftward[-1][0], active_info_rightward[-1][0] + 1)
    
    return active_range


import textwrap
import gzip

import pysam

import handygenome.tools as tools
import handygenome.refgenome.refgenome as refgenome
import handygenome.workflow as workflow
import handygenome.workflow.toolsetup as toolsetup
import handygenome.vcfeditor.indexing as indexing
import handygenome.vcfeditor.initvcf as initvcf
import handygenome.vcfeditor.headerhandler as headerhandler
import handygenome.variant.vcfspec as libvcfspec
import handygenome.variant.varianthandler as varianthandler


def argument_parser(cmdargs):
    def sanity_check(args):
        pass

    parser_dict = workflow.init_parser(
        description='Turns a Bismark *.cov.gz output file into a VCF.')

    workflow.add_infile_arg(
        parser_dict['required'], required=True,
        help=f'A Bismark output file which matched the pattern "*.cov.gz"')
    workflow.add_outfile_arg(parser_dict['required'], required=True, 
                             must_not_exist='ask')
    workflow.add_fasta_arg(parser_dict['required'], required=True)
    workflow.add_outfmt_arg(parser_dict['optional'], required=False)

    workflow.add_logging_args(parser_dict)
    workflow.add_index_arg(parser_dict)

    parser_dict['required'].add_argument(
        '--sampleid', required=True,
        help=f'The sample name for the input file.')

    args = parser_dict['main'].parse_args(cmdargs)
    sanity_check(args)

    return args


def add_meta(vcfheader):
    # ALT meta
    vcfheader.add_meta(
        key='ALT', items=[('ID', libvcfspec.CPGMET_ALT), 
                          ('Description', 'CpG site methylation')])

    # INFO meta
    headerhandler.addmeta_END(vcfheader)

    # FORMAT meta
    vcfheader.add_meta(
        key='FORMAT',
        items=[
            ('ID', 'nread_meth_bisulf_top'), ('Type', 'Integer'), 
            ('Number', 1),
            ('Description', ('The number of reads supporting cytosine '
                             'methylation, '
                             'belonging to the top(plus) strand from a '
                             'bisulfite sequencing result.'))])

    vcfheader.add_meta(
        key='FORMAT',
        items=[
            ('ID', 'nread_total_bisulf_top'), ('Type', 'Integer'), 
            ('Number', 1),
            ('Description', ('The number of all reads, '
                             'belonging to the top(plus) strand from a '
                             'bisulfite sequencing result.'))])

    vcfheader.add_meta(
        key='FORMAT',
        items=[
            ('ID', 'meth_fract_bisulf_top'), ('Type', 'Float'), 
            ('Number', 1),
            ('Description', ('The fraction of reads supporting cytosine '
                             'methylation, '
                             'belonging to the top(plus) strand from a '
                             'bisulfite sequencing result.'))])

    vcfheader.add_meta(
        key='FORMAT',
        items=[
            ('ID', 'nread_meth_bisulf_bot'), ('Type', 'Integer'), 
            ('Number', 1),
            ('Description', ('The number of reads supporting cytosine '
                             'methylation, '
                             'belonging to the bottom(minus) strand from a '
                             'bisulfite sequencing result.'))])

    vcfheader.add_meta(
        key='FORMAT',
        items=[
            ('ID', 'nread_total_bisulf_bot'), ('Type', 'Integer'), 
            ('Number', 1),
            ('Description', ('The number of all reads, '
                             'belonging to the bottom(minus) strand from a '
                             'bisulfite sequencing result.'))])

    vcfheader.add_meta(
        key='FORMAT',
        items=[
            ('ID', 'meth_fract_bisulf_bot'), ('Type', 'Float'), 
            ('Number', 1),
            ('Description', ('The fraction of reads supporting cytosine '
                             'methylation, '
                             'belonging to the bottom(minus) strand from a '
                             'bisulfite sequencing result.'))])


def init_vcfheader(chromdict, sampleid):
    vcfheader = initvcf.create_header(chromdict=chromdict, samples=[sampleid])
    add_meta(vcfheader)

    return vcfheader


def parse_linesp(linesp):
    assert linesp[1] == linesp[2], (
        f'Column 2 and Column 3 are different in this line: {linesp}')

    chrom = linesp[0]
    pos = int(linesp[1])
    meth_fract = 0.01 * float(linesp[3])
    nread_meth = int(linesp[4])
    nread_total = int(linesp[4]) + int(linesp[5])

    linesp_parsed = {'chrom': chrom, 'pos': pos, 'meth_fract': meth_fract,
                     'nread_meth': nread_meth, 'nread_total': nread_total}

    return linesp_parsed


def parse_one_line(infile):
    try:
        line = next(infile)
    except StopIteration:
        return None
    else:
        linesp = tools.get_linesp(line)
        linesp_parsed = parse_linesp(linesp)
        return linesp_parsed


def parse_two_lines(infile):
    linesp_parsed_prev = parse_one_line(infile)
    linesp_parsed = parse_one_line(infile)

    return linesp_parsed_prev, linesp_parsed


def handle_singleton(linesp_parsed, fasta, vcfheader, vr_buffer):
    if linesp_parsed['chrom'] in fasta.references:
        ref = fasta.fetch(linesp_parsed['chrom'], linesp_parsed['pos'] - 2, 
                          linesp_parsed['pos'] + 1)
        if ref[1] == 'C':
            if ref[2] == 'G':
                pos = linesp_parsed['pos']
                vr = vcfheader.new_record(start=(pos - 1))
                vr.contig = linesp_parsed['chrom']
                vr.ref = 'CG'
                vr.alts = ('<CPGMET>',)

                vr.samples[0]['nread_meth_bisulf_top'] = (
                    linesp_parsed['nread_meth'])
                vr.samples[0]['nread_total_bisulf_top'] = (
                    linesp_parsed['nread_total'])
                vr.samples[0]['meth_fract_bisulf_top'] = (
                    linesp_parsed['meth_fract'])
            else:
                vr = None
                #raise Exception(f'Unexpected REF sequence on a singleton line '
                                #f'position: "C" "nonG"\nline contents: {linesp}')
        elif ref[1] == 'G':
            if ref[0] == 'C':
                pos = linesp_parsed['pos'] - 1
                vr = vcfheader.new_record(start=(pos - 1))
                vr.contig = linesp_parsed['chrom']
                vr.ref = 'CG'
                vr.alts = ('<CPGMET>',)

                vr.samples[0]['nread_meth_bisulf_bot'] = (
                    linesp_parsed['nread_meth'])
                vr.samples[0]['nread_total_bisulf_bot'] = (
                    linesp_parsed['nread_total'])
                vr.samples[0]['meth_fract_bisulf_bot'] = (
                    linesp_parsed['meth_fract'])
            else:
                vr = None
                #raise Exception(f'Unexpected REF sequence on a singleton line '
                                #f'position: "nonC" "G"\n'
                                #f'line contents: {linesp_parsed}')
        else:
            vr = None
            #raise Exception(f'Unexpected REF sequence on a singleton line '
                            #f'position: neither C nor G'
                            #f'\nline contents: {linesp_parsed}')
    else:
        vr = None

    if vr is not None:
        vr_buffer.append(vr)


def handle_doublet(linesp_parsed_prev, linesp_parsed, vcfheader, fasta,
                   vr_buffer):
    if linesp_parsed['chrom'] in fasta.references:
        ref = fasta.fetch(linesp_parsed['chrom'], 
                          linesp_parsed_prev['pos'] - 1, linesp_parsed['pos'])
        if ref == 'CG':
            pos = linesp_parsed_prev['pos']
            vr = vcfheader.new_record(start=(pos-1))
            vr.contig = linesp_parsed['chrom']
            vr.ref = 'CG'
            vr.alts = ('<CPGMET>',)
            
            vr.samples[0]['nread_meth_bisulf_top'] = (
                linesp_parsed_prev['nread_meth'])
            vr.samples[0]['nread_total_bisulf_top'] = (
                linesp_parsed_prev['nread_total'])
            vr.samples[0]['meth_fract_bisulf_top'] = (
                linesp_parsed_prev['meth_fract'])

            vr.samples[0]['nread_meth_bisulf_bot'] = (
                linesp_parsed['nread_meth'])
            vr.samples[0]['nread_total_bisulf_bot'] = (
                linesp_parsed['nread_total'])
            vr.samples[0]['meth_fract_bisulf_bot'] = (
                linesp_parsed['meth_fract'])
        else:
            vr = None
    else:
        vr = None

    if vr is not None:
        vr_buffer.append(vr)


def check_linesp_adjacency(linesp_parsed_prev, linesp_parsed):
    return (
        (linesp_parsed_prev['pos'] == linesp_parsed['pos'] - 1) and
        (linesp_parsed_prev['chrom'] == linesp_parsed['chrom']))


def get_linesp_buffer(infile):
    linesp_parsed = parse_one_line(infile)
    if linesp_parsed is None:
        return None
    else:
        linesp_buffer = list()
        linesp_buffer.append(linesp_parsed)
        while True:
            linesp_parsed = parse_one_line(infile)
            if linesp_parsed is None:
                break
            else:
                linesp_buffer.append(linesp_parsed)
                if check_linesp_adjacency(linesp_buffer[-2], linesp_buffer[-1]):
                    break
                else:
                    continue

        return linesp_buffer


def get_vr_buffer(linesp_buffer, fasta, vcfheader):
    if linesp_buffer is None:
        return None
    else:
        vr_buffer = list()

        if len(linesp_buffer) == 1:
            handle_singleton(linesp_buffer[0], fasta, vcfheader, vr_buffer)
        elif len(linesp_buffer) >= 2:
            if check_linesp_adjacency(linesp_buffer[-2], linesp_buffer[-1]):
                for linesp_parsed in linesp_buffer[:-2]:
                    handle_singleton(linesp_parsed, fasta, vcfheader,
                                     vr_buffer)
                vr = handle_doublet(linesp_buffer[-2], linesp_buffer[-1],
                                    vcfheader, fasta, vr_buffer)
            else:
                for linesp_parsed in linesp_buffer:
                    handle_singleton(linesp_parsed, fasta, vcfheader,
                                     vr_buffer)

        return vr_buffer


def write_oufile(infile_path, outfile_path, mode_pysam, vcfheader, fasta, 
                 chromdict, logger):
    N_VR = 0
    out_vr_list = list()
    with gzip.open(infile_path, 'rt') as infile:
        while True:
            linesp_buffer = get_linesp_buffer(infile)
            vr_buffer = get_vr_buffer(linesp_buffer, fasta, vcfheader)

            if vr_buffer is None:
                # arrived at the end of the input file
                break
            else:
                N_VR += len(vr_buffer)
                for vr in vr_buffer:
                    out_vr_list.append(vr)

            if N_VR % 1_000_000 == 0:
                logger.info(
                    f'{N_VR:,} variant records processed. Current '
                    f'raw input line: {linesp_buffer[-1]}')

    # sort
    logger.info('Sorting variant records')
    out_vr_list.sort(key=varianthandler.get_vr_sortkey(chromdict))

    # write
    logger.info('Writing variant records')
    with pysam.VariantFile(outfile_path, mode=mode_pysam,
                           header=vcfheader) as out_vcf:
        for vr in out_vr_list:
            out_vcf.write(vr)


def main(cmdargs):
    args = argument_parser(cmdargs)
    logger = toolsetup.setup_logger(args)

    logger.info('BEGINNING')

    fasta = pysam.FastaFile(args.fasta_path)
    chromdict = refgenome.Chromdict.from_fasta(fasta)
    vcfheader = init_vcfheader(chromdict, args.sampleid)

    write_oufile(args.infile_path, args.outfile_path, args.mode_pysam, 
                 vcfheader, fasta, chromdict, logger)
                
    if not args.donot_index:
        indexing.index_vcf(args.outfile_path)

    logger.info('ALL SUCCESSFULLY FINISHED')




import os
import gzip
import multiprocessing
import itertools

import pysam
import numpy as np

import handygenome.tools as tools
import handygenome.refgenome.refgenome as refgenome
import handygenome.interval as libinterval
import handygenome.logutils as logutils
from handygenome.genomedf.genomedf import GenomeDataFrame as GDF


# pysam VariantFile mode related ones

"""
pysam file mode string

- Valid mode string pattern : ^[rwa]([bzu]?[0-9]?|[0-9]?[bzu]?)$
- Conflicting characters: (b,z,u)

- 'wb' : Compressed BCF (level 6), regardless of file name. 
- 'wb[0-9]' : BCF with level of compression indicated by the number, regardless of file name. 

- 'wz' : Compressed VCF (level 6), regardless of file name.
- 'wz[0-9]' : VCF with level of compression indicated by the number, regardless of file name. 

- 'wu' : Uncompressed VCF, regardless of file name.
- 'wu[0-9]' : Uncompressed VCF, regardless of file name.

- 'w[0-9]' : Uncompressed VCF, regardless of file name. 

- 'w' :
    *.vcf : uncompressed VCF
    *.vcf.gz : compressed VCF (level 6)
    *.bcf : compressed BCF (level 6)
    *.bcf.gz : compressed VCF (level 6)
"""

CYVCF2_FORMAT_DICT = {'v':'w', 'z':'wz', 'u':'wbu', 'b':'wb'}  
    # keys: bcftools format strings
    # values: cyvcf2 format strings
PYSAM_MODE_DICT = {'v':'wz0', 'z':'wz', 'u':'wb0', 'b':'wb'}
    # keys: bcftools format strings
    # values: pysam format strings
#DEFAULT_MODE_BCFTOOLS = 'z'


def write_mode_arghandler(mode_bcftools, mode_pysam):
    """
    Args:
        mode_bcftools: bcftools style mode string
        mode_pysam: pysam style mode string

    If 'mode_pysam' is set as a non-None value,
    it overrides 'mode_bcftools'.
    """
    if mode_pysam is None:
        return PYSAM_MODE_DICT[mode_bcftools]
    else:
        return mode_pysam

###

def get_vcf_format(vcf_path):
    with pysam.VariantFile(vcf_path) as vcf:
        is_vcf = vcf.is_vcf
        #is_bcf = vcf.is_bcf
        comp = vcf.compression

    if comp not in ('NONE', 'GZIP', 'BGZF'):
        raise Exception(f'Unexpected compression type')

    is_bgzf = (comp == 'BGZF')

    return is_vcf, comp, is_bgzf


def check_has_index(vcf_path):
    with pysam.VariantFile(vcf_path) as vcf:
        has_index = (vcf.index is not None)
    return has_index


def get_indexfile_path(vcf_path):
    if os.path.exists(f'{vcf_path}.csi'):
        return f'{vcf_path}.csi'
    elif os.path.exists(f'{vcf_path}.tbi'):
        return f'{vcf_path}.tbi'
    else:
        return None


def make_index(vcf_path):
    index_filename = f'{vcf_path}.csi'
    pysam.tabix_index(vcf_path, preset='vcf', csi=True, index=index_filename)
    return index_filename


###################################################
# VCF position extraction and fetch - new version #
###################################################

'''pysam.VariantFile.fetch regards span of a VCF record as an interval:
start0 = POS - 1
end0 = start0 + len(REF)
'''

def get_vcf_positions_new(vcf_path, split=None, nproc=1):
    if split is None:
        split = nproc * 10

    refver = refgenome.infer_refver_vcfpath(vcf_path)
    if not check_has_index(vcf_path):
        make_index(vcf_path) 

    allregion_gdf = GDF.all_regions(refver, assembled_only=False)
    allregion_split = allregion_gdf.equal_length_split(n=split)

    with multiprocessing.Pool(nproc) as pool:
        args = (
            (vcf_path, gdf) 
            for gdf in allregion_split
        )
        pool_result = pool.starmap(
            get_vcf_positions_new_target,
            args,
        )

    all_vr_intervals = list(itertools.chain.from_iterable(pool_result))
    chroms, start0s, end0s = zip(*all_vr_intervals)
    result_gdf = GDF.from_data(refver=refver, chroms=chroms, start0s=start0s, end0s=end0s)
    result_gdf = result_gdf.merge()

    return result_gdf


def get_vcf_positions_new_target(vcf_path, gdf):
    result = list()
    with pysam.VariantFile(vcf_path) as vcf:
        for chrom, start0, end0 in gdf.iter_coords():
            for vr in vcf.fetch(chrom, start0, end0):
                result.append(
                    (vr.chrom, vr.pos - 1, vr.pos - 1 + len(vr.ref))
                )
    return result


def get_vcf_fetchregions_new(vcf_path, n, nproc=1):
    refver = refgenome.infer_refver_vcfpath(vcf_path)
    position_gdf = get_vcf_positions_new(vcf_path, nproc=nproc)
    fetchregion_gdfs = list()
    for subgdf in position_gdf.equal_nrow_split(n=n):
        chrom_left = subgdf.chromosomes[0]
        start0_left = subgdf.starts[0]
        chrom_right = subgdf.chromosomes[-1]
        end0_right = subgdf.ends[-1]
        fetchregion_gdfs.append(
            GDF.from_margins(
                refver, chrom_left, start0_left, chrom_right, end0_right,
            )
        )

    return fetchregion_gdfs


def get_vr_fetcher_new(vcf, refver, chrom=None, start0=None, end0=None):
    if chrom is None:
        for vr in vcf.fetch():
            yield vr
    else:
        if start0 is None:
            start0 = 0
        if end0 is None:
            end0 = refgenome.get_chromdict(refver)[chrom]
        
        for vr in vcf.fetch(contig=chrom, start=start0, stop=end0):
            yield vr


###################################################
# VCF position extraction and fetch - old version #
###################################################

def get_vcf_positions(vcf_path, as_iter=False, verbose=False):
    """Returns:
        as_iter==True: A list of tuples (chrom, start0, end0) of each variant record
        as_iter==False: np.ndarray
    """
    is_vcf, comp, is_bgzf = get_vcf_format(vcf_path)
    if is_vcf:
        line_iter = get_line_iter(vcf_path, comp, verbose)
        result_iter = parse_line_iter(line_iter)
    else:  # BCF
        result_iter = get_vcf_positions_pysam(vcf_path)

    if as_iter:
        return result_iter
    else:
        #return list(result_iter)
        return np.array(
            list(result_iter), 
            dtype=[('Chromosome', object), ('Start', int), ('End', int)],
        )


def split_vcf_positions(vcf_positions, n):
    """Splits such that adjacent sub-positions do not meet each other with
    identical POS values
    """
    diffresult = np.diff(vcf_positions['Start'])
    nondiff_indexes = np.where(diffresult == 0)[0] + 1
    spittable_indexes = np.setdiff1d(np.arange(vcf_positions.shape[0]), nondiff_indexes)
    spittable_indexes_split = [
        x for x in np.array_split(spittable_indexes, n) 
        if x.shape[0] != 0
    ]
    splitpoints = [x[0] for x in spittable_indexes_split if x[0] != 0]
    split_positions = np.split(vcf_positions, splitpoints)
        # this works even when "splitpoints" is an empty list

    return split_positions


def get_fetchargs_from_vcf_positions(vcf_positions, refver):
    chrom_left = vcf_positions[0]['Chromosome']
    start0_left = vcf_positions[0]['Start']
    chrom_right = vcf_positions[-1]['Chromosome']
    end0_right = vcf_positions[-1]['Start'] + 1  
        # only the 1-base position indicated by POS value is used
    chromdict = refgenome.get_chromdict(refver)

    intvlist = libinterval.IntervalList.from_margin(
        chromdict, chrom_left, start0_left, chrom_right, end0_right,
    )
    fetchargs_list = [
        (intv.chrom, intv.start0, intv.end0) for intv in intvlist
    ]
    return fetchargs_list


def get_vcf_fetchregions(vcf_path, n, refver, verbose=False):
    all_position_info = get_vcf_positions(vcf_path, as_iter=False, verbose=verbose)
    split_position_info = split_vcf_positions(all_position_info, n)
    result = [
        get_fetchargs_from_vcf_positions(position_info, refver)
        for position_info in split_position_info
    ]

    return result


def get_vr_fetcher(vcf, refver, chrom=None, start0=None, end0=None, respect_refregion=False):
    if chrom is None:
        for vr in vcf.fetch():
            yield vr
    else:
        if start0 is None:
            start0 = 0
        if end0 is None:
            end0 = refgenome.get_chromdict(refver)[chrom]
        
        fetcher = vcf.fetch(contig=chrom, start=start0, stop=end0)

        if respect_refregion:
            def check_vr_inclusion(vr, chrom, start0, end0):
                vr_start0 = vr.pos - 1
                vr_end0 = vr_start0 + len(vr.ref)
                return (
                    (vr.contig == chrom)
                    and (vr_end0 > start0)
                    and (vr_start0 < end0)
                )
        else:
            def check_vr_inclusion(vr, chrom, start0, end0):
                vr_start0 = vr.pos - 1
                vr_end0 = vr.pos
                return (
                    (vr.contig == chrom)
                    and (vr_end0 > start0)
                    and (vr_start0 < end0)
                )

        for vr in fetcher:
            if check_vr_inclusion(vr, chrom, start0, end0):
                yield vr

#################
# miscellaneous #
#################

def get_vcf_lineno(vcf_path, verbose=False):
    is_vcf, comp, is_bgzf = get_vcf_format(vcf_path)
    if is_vcf:
        line_iter = get_line_iter(vcf_path, comp, verbose)
        return count_iterator(line_iter)
    else:  # BCF
        return get_vcf_lineno_pysam(vcf_path)


def count_iterator(iterator):
    idx = -1
    for idx, x in enumerate(iterator):
        pass
    return idx + 1


def get_line_iter(vcf_path, comp, verbose):
    if comp in ('BGZF', 'GZIP'):
        raw_iter = gzip.open(vcf_path, 'rb')
    elif comp == 'NONE':
        raw_iter = open(vcf_path, 'rb')

    if verbose:
        for idx, line in enumerate(raw_iter):
            if idx % 100000 == 0:
                print(idx)
            if not line.startswith(b'#'):
                yield line

    else:
        for line in raw_iter:
            if not line.startswith(b'#'):
                yield line


def parse_line_iter(line_iter):
    """Accepts byte mode iterator"""
    for line in line_iter:
        linesp = line.split(b'\t')
        pos0 = int(linesp[1]) - 1
        yield (linesp[0].decode(), pos0, pos0 + len(linesp[3]))


def get_vcf_positions_pysam(vcf_path):
    with pysam.VariantFile(vcf_path) as vcf:
        for vr in vcf.fetch():
            yield (vr.contig, vr.pos - 1, vr.pos - 1 + len(vr.ref))


def get_vcf_lineno_pysam(vcf_path):
    with pysam.VariantFile(vcf_path) as vcf:
        lineno = count_iterator(vcf.fetch())
    return lineno


# this does not work
def get_vcf_noerr(*args, **kwargs):
    with contextlib.redirect_stderr(io.StringIO()) as err, \
            contextlib.redirect_stdout(io.StringIO()) as out:
        vcf = pysam.VariantFile(*args, **kwargs)

    for buf in (err, out):
        msg = buf.getvalue()
        if not msg.startswith('[E::idx_find_and_load] '
                              'Could not retrieve index file for'):
            print(msg, end='', file=sys.stderr)

    return vcf


